import phenopackets as PPKt

import pandas as pd
from typing import Optional
from .age_isoformater import AgeIsoFormater
from .constants import Constants


class AgeOfDeathColumnMapper:

    def __init__(self, column_name, string_to_iso_d=None) -> None:
        """
        :param column_name: Name of the Age of death column in the original table
        :type column_name: str
        :param string_to_iso_d: dictionary from free text (input table) to ISO8601 strings
        :type string_to_iso_d: Dict[str,str], optional
        """

        if string_to_iso_d is None:
            string_to_iso_d = {}
        if column_name is None:
            raise ValueError("Must provide non-null column_name argument")
        self._column_name = column_name
        self._string_to_iso_d = string_to_iso_d

    def map_cell_to_vital_status(self, cell_contents) -> Optional[PPKt.VitalStatus]:

        """
        Map a single cell of the table

        :param cell_contents: The text contained in a single cell of the table
        :type cell_contents: can be a string or numerical type
        """

        contents = str(cell_contents)
        if contents not in self._string_to_iso_d:
            return None
        iso_age = self._string_to_iso_d.get(contents)
        vstatus = PPKt.VitalStatus()
        vstatus.status = PPKt.VitalStatus.DECEASED
        vstatus.time_of_death.age.iso8601duration = iso_age
        return vstatus

    @property
    def column_name(self):
        return self._column_name



import unittest
import os
import re
from src.pyphetools.creation import AgeColumnMapper

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')
# The API requires us to pass a column name but the column name will not be used in the tests
TEST_COLUMN = "test"

string_to_iso_dict = {
    'Fetus': 'nan',
    "1.5": "P1Y6M",
    "3.5": "P3Y6M",
    'birth': "P1D",
    '51 days': "P1M21D",
    '19 months': "P1Y7M",
    '10 monhts': "P10M",
    '14 months': "P1Y2M",
    '23 months': "P1Y11M",
    '22 mohts': "P1Y10M",
    '9 months': "P9M",
    '12 years': "P12Y",
    '1 day of age': "P1D",
    '4.5 years': "P4Y6M",
    '4 months': "P4M",
    '1 month': "P1M"
}

class TestOptionMapper(unittest.TestCase):

    def test_year(self):
        ageMapper = AgeColumnMapper.by_year(column_name=TEST_COLUMN)
        p3 = ageMapper.map_cell("3")
        self.assertEqual("P3Y", p3)
        p42 = ageMapper.map_cell("42")
        self.assertEqual("P42Y", p42)

    def test_iso8601(self):
        ageMapper = AgeColumnMapper.iso8601(column_name=TEST_COLUMN)
        p3y = ageMapper.map_cell("P3Y")
        self.assertEqual("P3Y", p3y)
        p3m25d = ageMapper.map_cell("P3Y2M5D")
        self.assertEqual("P3Y2M5D", p3m25d)

    def test_year_month_both_1(self):
        ageMapper = AgeColumnMapper.by_year_and_month(column_name=TEST_COLUMN)
        age_string = "14 y 8 m"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P14Y8M", age_iso)

    def test_year_month_both_2(self):
        ageMapper = AgeColumnMapper.by_year_and_month(column_name=TEST_COLUMN)
        age_string = "7 y 6 m"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P7Y6M", age_iso)

    def test_year_month_both_3(self):
        ageMapper = AgeColumnMapper.by_year_and_month(column_name=TEST_COLUMN)
        age_string = "7y6m"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P7Y6M", age_iso)

    def test_year_month_just_year_1(self):
        ageMapper = AgeColumnMapper.by_year_and_month(column_name=TEST_COLUMN)
        age_string = "7 y"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P7Y", age_iso)

    def test_year_month_just_month_1(self):
        ageMapper = AgeColumnMapper.by_year_and_month(column_name=TEST_COLUMN)
        age_string = "2 m"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P2M", age_iso)

    def test_int_or_float_regex(self):
        int_or_float = r"(\d+)(\.\d+)?"
        p=re.compile(int_or_float)
        y_only = "42"
        results = p.search(y_only).groups()
        self.assertEqual(2, len(results))
        self.assertEqual('42', results[0])
        self.assertIsNone(results[1])
        y_and_m = "42.25"
        results = p.search(y_and_m).groups()
        self.assertEqual(2, len(results))
        self.assertEqual('42', results[0])
        self.assertEqual('.25', results[1])

    def test_fractional_year_strings(self):
        ageMapper = AgeColumnMapper.by_year(column_name=TEST_COLUMN)
        age_string = "2"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P2Y", age_iso)
        age_string = "4.75"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P4Y9M", age_iso)
        age_string = "5.9"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P5Y11M", age_iso)
        age_string = "6.25"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P6Y3M", age_iso)
        age_string = "8.1"
        age_iso = ageMapper.map_cell(age_string)
        self.assertEqual("P8Y1M", age_iso)

    def test_custom_dictionary(self):
        ageMapper = AgeColumnMapper.custom_dictionary(column_name=TEST_COLUMN, string_to_iso_d=string_to_iso_dict)
        age_iso = ageMapper.map_cell("1.5")
        self.assertEqual("P1Y6M", age_iso)
        age_iso = ageMapper.map_cell("3.5")
        self.assertEqual("P3Y6M", age_iso)
        age_iso = ageMapper.map_cell("birth")
        self.assertEqual("P1D", age_iso)
        age_iso = ageMapper.map_cell("NOT THERE")
        NOT_PROVIDED = 'NOT_PROVIDED' # from Constants.py which is not exported
        self.assertEqual(NOT_PROVIDED, age_iso)

    def test_by_month(self):
        ageMapper = AgeColumnMapper.by_month(column_name=TEST_COLUMN)
        age_iso = ageMapper.map_cell(5)
        self.assertEqual("P5M", age_iso)
        age_iso = ageMapper.map_cell("5")
        self.assertEqual("P5M", age_iso)
        age_iso = ageMapper.map_cell(0.5)
        self.assertEqual("P15D", age_iso)
        age_iso = ageMapper.map_cell(0.8)
        self.assertEqual("P24D", age_iso)
        age_iso = ageMapper.map_cell(12)
        self.assertEqual("P1Y", age_iso)





from typing import List
import pandas as pd
from .column_mapper import ColumnMapper
from .hp_term import HpTerm


class ConstantColumnMapper(ColumnMapper):
    """Column mapper for cases in which all patients have an (optionally excluded) HPO term.
    :param column_name: name of the column in the pandas DataFrame
    :type column_name: str
    :param hpo_id: HPO  id, e.g., HP:0004321
    :type hpo_id: str
    :param hpo_label: Corresponding term label
    :type hpo_label: str
    :param term_list: list of lists with [label, hpo_id
    :type term_list: List[lst]
    :param excluded: if True, then all individuals had this feature explicitly excluded
    :type excluded: bool
    """
    def __init__(self, column_name,  hpo_id=None, hpo_label=None, term_list=None, excluded:bool=False) -> None:

        super().__init__(column_name=column_name)
        self._hpo_id = hpo_id
        if hpo_id is None and hpo_label is None and term_list is not None:
            self._hpo_terms = []
            for term in term_list:
                if excluded:
                    hpoterm = HpTerm(label=term[0], hpo_id=term[1], observed=False)
                else:
                    hpoterm = HpTerm(label=term[0], hpo_id=term[1], observed=True)
                self._hpo_terms.append(hpoterm)
        elif term_list is None and hpo_id is not None and hpo_label is not None:
            if excluded:
                hpoterm = HpTerm(label=hpo_label, hpo_id=hpo_id, observed=False)
            else:
                hpoterm = HpTerm(label=hpo_label, hpo_id=hpo_id, observed=True)
            self._hpo_terms = [hpoterm]
        else:
            raise ValueError(f"Error: Either hpo_id and hpo_label are not not or a list of HPO terms is passed")
        self._excluded = excluded

    def map_cell(self, cell_contents) -> List[HpTerm]:
        """if this mapper is used, then all individuals in the table have the list of HPO terms

        Args:
            cell_contents (str): not used, can be None or any other value

        Returns:
            List[HpTerm]: list of HPO terms
        """
        return self._hpo_terms

    def preview_column(self, df:pd.DataFrame) -> pd.DataFrame:
        if not isinstance(df, pd.DataFrame):
            raise ValueError("df argument must be pandas DataFrame, but was {type(column)}")
        dlist = []
        column = df[self._column_name]
        for _, value in column.items():
            display = ";".join(hpterm.display_value for hpterm in self._hpo_terms)
            dlist.append(display)
        return pd.DataFrame(dlist)




from collections import defaultdict
import phenopackets as PPKt
from typing import Dict, List, Set
import sys

from ..creation.constants import Constants
from ..creation import Individual, HpTerm, MetaData
from .simple_patient import SimplePatient
from .html_table_generator import HtmlTableGenerator


#


class Age2Day:
    """
    Convenience function to help sort ages but converting ISO8601 strings into the number of days

        sorted_age = sorted(l, key=lambda x: x.days)

    following this, retrieve the original string as x.key
    """
    def __init__(self, age, days) -> None:
        self.key = age
        self.days = days


#@DeprecationWarning("This class will be replaced by IndividualTable and will be deleted in a future version")
class PhenopacketTable:
    """
    This class creates a table with a summary of all phenopackets in a cohort of individuals
    Create Individual objects and transform them into phenopackets, or import GA4GH phenopackets and display them.

        from IPython.display import HTML, display
        phenopackets = [i.to_ga4gh_phenopacket(metadata=metadata) for i in individuals]
        table = PhenopacketTable.from_phenopackets(phenopacket_list=phenopackets)
        display(HTML(table.to_html()))

    Alternatively,

        table = PhenopacketTable.from_individuals(individual_list=individuals, metadata=metadata)
        display(HTML(table.to_html()))
    """
    def __init__(self, phenopacket_list:List[PPKt.Phenopacket]=None,
                        individual_list:List[Individual]=None,
                        metadata:MetaData=None) -> None:
        """
        :param phenopacket_list: List of GA4GH phenopackets to be displayed
        :type phenopacket_list: List[PPKt.Phenopacket]
        :param individual_list: List of Indidivual objects to be displayed
        :type individual_list: List[Individual]
        :param metadata: metadata about the individuals
        :type metadata: MetaData
        """
        if phenopacket_list is None and individual_list is not None and metadata is not None:
            self._phenopacket_list =  [i.to_ga4gh_phenopacket(metadata=metadata.to_ga4gh()) for i in individual_list]
        elif phenopacket_list is not None and individual_list is None:
            if metadata is not None:
                raise ValueError("metadata argument not needed for phenopacket list")
            self._phenopacket_list = phenopacket_list

    @staticmethod
    def from_phenopackets(phenopacket_list:List[PPKt.Phenopacket]):
        """Initialize PhenopacketTable from list of GA4GH Phenopackets

        :param phenopacket_list: list of GA4GH Phenopackets
        :type phenopacket_list: List[PPKt.Phenopacket]

        :returns: PhenopacketTable for displaying information about a cohort
        :rtype: PhenopacketTable
        """
        if not isinstance(phenopacket_list, list):
            raise ValueError(f"Expecting a list but got {type(phenopacket_list)}")
        if len(phenopacket_list) == 0:
            raise ValueError("phenopacket_list was empty")
        ppkt = phenopacket_list[0]
        if not isinstance(ppkt, PPKt.Phenopacket):
            raise ValueError(f"phenopacket_list argument must be list of Phenopacket objects but was {type(ppkt)}")
        return PhenopacketTable(phenopacket_list=phenopacket_list)

    @staticmethod
    def from_individuals(individual_list:List[Individual], metadata:MetaData):
        """Initialize PhenopacketTable from list of GA4GH Phenopackets

        :param phenopacket_list: list of GA4GH Phenopackets
        :type phenopacket_list: List[PPKt.Phenopacket]

        :returns: PhenopacketTable for displaying information about a cohort
        :rtype: PhenopacketTable
        """
        if not isinstance(individual_list, list):
            raise ValueError(f"Expecting a list but got {type(individual_list)}")
        if len(individual_list) == 0:
            raise ValueError("individual_list was empty")
        indi = individual_list[0]
        if not isinstance(indi, Individual):
            raise ValueError(f"individual_list argument must be list of Individual objects but was {type(indi)}")
        if not isinstance(metadata, MetaData):
            raise ValueError(f"metadata argument must be pyphetools MetaData object but was {type(metadata)}")
        return PhenopacketTable(individual_list=individual_list, metadata=MetaData)

    def _simple_patient_to_table_row(self, spat:SimplePatient) -> List[str]:
        """
        private method intended to create one table row that represents one individual
        :param spat: An object that represents one individual
        :type spat: SimplePatient
        """
        row_items = []
        # Patient information
        pat_info = spat.get_subject_id() + " (" + spat.get_sex() + "; " + spat.get_age() + ")"
        row_items.append( pat_info)
        row_items.append( spat.get_disease())
        # Variant information
        var_list = spat.get_variant_list()
        if len(var_list) == 0:
            row_items.append("n/a" )
        elif len(var_list) == 1:
            var = var_list[0]
            row_items.append( var.get_display() )
        else:
            cell_items = []
            cell_items.append("<ul>")
            for var in var_list:
                cell_items.append("<li>" + var.get_display() + "</li>")
            cell_items.append("</ul>")
            row_items.append( " ".join(cell_items) )
        # HPO information
        hpo_html = self.get_hpo_cell(spat.get_term_by_age_dict())
        row_items.append( hpo_html )
        return row_items


    def get_hpo_cell(self, term_by_age_dict:Dict[str,HpTerm]) -> str:
        """
        :param term_by_age_dict: A dictionary with key - ISO8601 string, value - list of HpTerm objects
        :type term_by_age_dict: Dict[str,HpTerm]
        :returns: HTML code for the HTML cell
        :rtype: str
        """
        lines = []
        age2day_list = PhenopacketTable.get_sorted_age2data_list(term_by_age_dict.keys())
        sorted_age = sorted(age2day_list, key=lambda x: x.days)
        for onset in sorted_age:
            hpo_list = term_by_age_dict.get(onset.key)
            hpos = "; ".join([hpo.__str__() for hpo in hpo_list])
            if onset.key == Constants.NOT_PROVIDED:
                lines.append(hpos)
            else:
                lines.append(f"<b>{onset.key}</b>: {hpos}")
        return "<br/>".join(lines)

    def to_html(self):
        """create an HTML table with patient ID, age, sex, genotypes, and PhenotypicFeatures
        """
        ppack_list = self._phenopacket_list
        spat_list = []
        pmid_count_d = defaultdict(int)
        no_pmid_found = 0
        pmid_found = 0
        for pp in ppack_list:
            spat = SimplePatient(ga4gh_phenopacket=pp)
            if spat.has_pmid():
                pmid_count_d[spat.get_pmid()] += 1
                pmid_found += 1
            else:
                no_pmid_found += 1
            spat_list.append(spat)
        # Create caption
        if pmid_found == 0:
            capt = f"{len(ppack_list)} phenopackets - no PMIDs (consider adding this information to the MetaData)"
        else:
            pmid_strings = []
            for k, v in pmid_count_d.items():
                pmid_strings.append(f"{k} (n={v})")
            pmid_str = "; ".join(pmid_strings)
            n_phenopackets = len(ppack_list)
            if n_phenopackets == 1:
                capt = f"{n_phenopackets} phenopacket - {pmid_str}"
            else:
                capt = f"{n_phenopackets} phenopackets - {pmid_str}"
        header_items = ["Individual", "Disease", "Genotype", "Phenotypic features"]
        rows = []
        for spat in spat_list:
            rows.append(self._simple_patient_to_table_row(spat))
        generator = HtmlTableGenerator(caption=capt, header_items=header_items, rows=rows)
        return generator.get_html()

    @staticmethod
    def iso_to_days(iso_age:str) -> int:
        """
        Transform the ISO8601 age strings (e.g., P3Y2M) into the corresponding number of days to facilitate sorting.

        Note that if age is not provided we want to sort it to the end of the list so we transform to a very high number of days.

        :param iso_age: ISO8601 age string (e.g., P3Y2M)
        :type iso_age: str
        :returns: number of days
        :rtype: int
        """
        if iso_age == Constants.NOT_PROVIDED:
            days = sys.maxsize
        elif not iso_age.startswith("P"):
            raise ValueError(f"Invlaid age string: {age}")
        else:
            days = 0
            age = iso_age[1:]
            N = len(age)
            y = age.find("Y")
            if y != -1:
                days = days + int(365.25*int(age[:y]))
                age = age[y+1:]
            m = age.find("M")
            if m != -1:
                days = days + int(30.436875*int(age[:m]))
                age = age[m+1:]
            d = age.find("D")
            if d != -1:
                days = days + int(age[:d])
        return days

    @staticmethod
    def get_sorted_age2data_list(ages:Set[str]) -> List[Age2Day]:
        """
        Create a sorted list of Age2Day objects that we use to display the age in the HTML output.

        :param ages: A set of ISO 8601 age strings
        :type ages: Set[str]
        :returns: A list of sorted Age2Day objects
        :rtype:  List[Age2Day]
        """
        age2day_list = list(Age2Day(age, PhenopacketTable.iso_to_days(age)) for age in ages)
        sorted_list = sorted(age2day_list, key=lambda x: x.days)
        return sorted_list






import requests
from .hgvs_variant import HgvsVariant


URL_SCHEME = "https://rest.variantvalidator.org/VariantValidator/variantvalidator/%s/%s%%3A%s/%s?content-type=application%%2Fjson"
ACCEPTABLE_GENOMES = {"GRCh37", "GRCh38", "hg19", "hg38"}



class VariantValidator:
    """
    Check and Encode HGVS string

    The default mRNA transcript for the HGVS strings should be added in the constructor byt can be overridden in the
    encode_hgvs method.

    :param genome_build: The genome assembly, one of "GRCh37", "GRCh38", "hg19", "hg38"
    :type genome_build: str
    :param transcript: An mRNA transcript that is the reference for the HGVS string, opt
    :type transcript: str
    """

    def __init__(self, genome_build, transcript=None):
        """
        Constructor
        """
        if genome_build not in ACCEPTABLE_GENOMES:
            raise ValueError(f"genome_build \"{genome_build}\" not recognized")
        self._genome_assembly = genome_build
        self._transcript = transcript

    def encode_hgvs(self, hgvs, custom_transcript=None):
        """
        Encode an HGVS string as a pyphetools Variant object

        :param hgvs: Human Genome Variation Society (HGVS) representation of a variant, e.g., c.36613706dup
        :type hgvs: str
        :param custom_transcript: a transcript (e.g., NM_001848.2), if different from the default transcript, optional
        :type custom_transcript: str
        :returns: pyphetools Variant object
        :rtype: HgvsVariant
        """
        if custom_transcript is not None:
            transcript = custom_transcript
        elif self._transcript is not None:
            transcript = self._transcript
        else:
            raise ValueError("Cannot run variant validator without transcript")
        api_url = URL_SCHEME % (self._genome_assembly, transcript, hgvs, transcript)
        #f"https://rest.variantvalidator.org/VariantValidator/variantvalidator/{self._genome_assembly}/{transcript}%3A{hgvs}/{transcript}"
        #
        print(api_url)
        response = requests.get(api_url)
        # We expect to get a dictionary with three keys. The first is the name of the variant, e.g., ACC:HGVS, then we
        # get flag and metadata
        vv_dict = response.json()
        if 'flag' in vv_dict:
            if vv_dict['flag'] != 'gene_variant':
                flag = vv_dict['flag']
                raise ValueError(f"Expecting to get a gene_variant from Variant Validator but got {flag}")
        # This is the easiest way to get the variant key, which may contain difficult characters
        variant_key = [k for k in vv_dict.keys() if k not in {'flag', 'metadata'}][0]
        var = vv_dict[variant_key]
        hgnc = None
        if 'gene_ids' in var and 'hgnc_id' in var['gene_ids']:
            hgnc = var['gene_ids']['hgnc_id']
        symbol = None
        if 'gene_symbol' in var:
            symbol = var['gene_symbol']
        assemblies = var['primary_assembly_loci']
        if not self._genome_assembly in assemblies:
            raise ValueError(f"Could not identified {self._genome_assembly} in Variant Validator response")
        assembly = assemblies[self._genome_assembly]
        hgvs_transcript_var = var.get('hgvs_transcript_variant', None)
        genomic_hgvs = assembly.get('hgvs_genomic_description', None)
        reference_sequence_records = var.get('reference_sequence_records', None)
        if reference_sequence_records is not None:
            transcript = reference_sequence_records['transcript']
            if transcript.startswith('https://www.ncbi.nlm.nih.gov/nuccore/'):
                transcript = transcript[37:]
        else:
            transcript = None
        # 'vcf': {'alt': 'C', 'chr': '16', 'pos': '1756403', 'ref': 'CG'}},
        if not 'vcf' in assembly:
            raise ValueError(f"Could not identify vcf element in Variant Validator genome assembly response")
        return HgvsVariant(assembly=self._genome_assembly, vcf_d=assembly['vcf'], symbol=symbol,
                        hgnc=hgnc, transcript=transcript, hgvs=hgvs_transcript_var, g_hgvs=genomic_hgvs)





import unittest
from src.pyphetools.visualization import SimpleAge


class TestSimpleAge(unittest.TestCase):

    def test_1month_1week(self):
        sa = SimpleAge(age_string="P1M1W")
        onset_term = sa.to_hpo_onset_term()
        self.assertEqual("HP:0003593", onset_term.id)
        self.assertEqual("Infantile onset", onset_term.label)

    def test_1week(self):
        sa = SimpleAge(age_string="P1W")
        onset_term = sa.to_hpo_onset_term()
        self.assertEqual("HP:0003623", onset_term.id)
        self.assertEqual("Neonatal onset", onset_term.label)

    def test_1day(self):
        sa = SimpleAge(age_string="P1D")
        onset_term = sa.to_hpo_onset_term()
        self.assertEqual("HP:0003577", onset_term.id)
        self.assertEqual("Congenital onset", onset_term.label)

    def test_0day(self):
        sa = SimpleAge(age_string="P0D")
        onset_term = sa.to_hpo_onset_term()
        self.assertEqual("HP:0003577", onset_term.id)
        self.assertEqual("Congenital onset", onset_term.label)

    def test_2Y3M1D(self):
        sa = SimpleAge(age_string="P2Y3M1D")
        onset_term = sa.to_hpo_onset_term()
        self.assertEqual("HP:0011463", onset_term.id)
        self.assertEqual("Childhood onset", onset_term.label)

    def test_11y(self):
        sa = SimpleAge(age_string="P11Y1D")
        onset_term = sa.to_hpo_onset_term()
        self.assertEqual("HP:0003621", onset_term.id)
        self.assertEqual("Juvenile onset", onset_term.label)

    def test_18y(self):
        sa = SimpleAge(age_string="P18Y2M1D")
        onset_term = sa.to_hpo_onset_term()
        self.assertEqual("HP:0011462", onset_term.id)
        self.assertEqual("Early young adult onset", onset_term.label)

    def test_22y(self):
        sa = SimpleAge(age_string="P22Y7M1D")
        onset_term = sa.to_hpo_onset_term()
        self.assertEqual("HP:0025709", onset_term.id)
        self.assertEqual("Intermediate young adult onset", onset_term.label)


    def test_39y(self):
        sa = SimpleAge(age_string="P39Y7M1D")
        onset_term = sa.to_hpo_onset_term()
        self.assertEqual("HP:0025710", onset_term.id)
        self.assertEqual("Late young adult onset", onset_term.label)

    def test_mideval(self):
        sa = SimpleAge(age_string="P45Y7M1D")
        onset_term = sa.to_hpo_onset_term()
        self.assertEqual("HP:0003596", onset_term.id)
        self.assertEqual("Middle age onset", onset_term.label)

    def test_late(self):
        sa = SimpleAge(age_string="P95Y7M1D")
        onset_term = sa.to_hpo_onset_term()
        self.assertEqual("HP:0003584", onset_term.id)
        self.assertEqual("Late onset", onset_term.label)






import abc
from typing import List

from .hp_term import HpTerm


class HpoConceptRecognizer(metaclass=abc.ABCMeta):
    """
    This abstract class acts as an interface for classes that implement parse_cell to perform HPO-based concept recognition.
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def parse_cell(self, cell_contents, custom_d=None) -> List[HpTerm]:
        """
        parse HPO Terms from the contents of a cell of the original table

        :param cell_contents: a cell of the original table
        :type cell_contents: str
        :param custom_d: a dictionary with keys for strings in the original table and their mappings to HPO labels
        :type custom_d: Dict[str,str], optional
        """
        pass


    @abc.abstractmethod
    def parse_cell_for_exact_matches(self, cell_contents, custom_d) -> List[HpTerm]:
        """
        Identify HPO Terms from the contents of a cell whose label exactly matches a string in the custom dictionary

        :param cell_contents: a cell of the original table
        :type cell_contents: str
        :param custom_d: a dictionary with keys for strings in the original table and their mappings to HPO labels
        :type custom_d: Dict[str,str]
        """
        pass

    @abc.abstractmethod
    def get_term_from_id(self, hpo_id) -> HpTerm:
        """
        :param hpo_id: an HPO identifier, e.g., HP:0004372
        :type hpo_id: str
        :returns: corresponding HPO term
        :rtype: HpTerm
        """
        pass

    @abc.abstractmethod
    def get_term_from_label(self, label) -> HpTerm:
        """
        :param label: an HPO label, e.g., Arachnodactyly
        :type label: str
        :returns: corresponding HPO term
        :rtype: HpTerm
        """
        pass

    @abc.abstractmethod
    def initialize_simple_column_maps(self, column_name_to_hpo_label_map, observed, excluded, non_measured=None):
        """
        Create a dictionary of SimpleColumnMappers
        """
        pass

    @abc.abstractmethod
    def get_hpo_ontology(self):
        """
        :returns: a reference to the HPO-toolkit Ontology object for the HPO
        :rtype: hpotk.MinimalOntology
        """
        pass



import unittest
from src.pyphetools.creation import StructuralVariant


class TestVariantValidator(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        variant = "46,XY.ish del(7)(p14.1)(RP11-816F16-)"
        genotype = "heterozygous"
        gene = "GLI3"
        gene_id = "HGNC:4319"
        sv = StructuralVariant.chromosomal_deletion(cell_contents=variant,
                                                    gene_symbol=gene,
                                                    gene_id=gene_id)
        sv.set_heterozygous()
        cls._variant_interpretation = sv.to_ga4gh_variant_interpretation()

    def test_label(self):
        var_inter = self._variant_interpretation
        variant_descriptor = var_inter.variation_descriptor
        label = variant_descriptor.label
        self.assertEqual("46,XY.ish del(7)(p14.1)(RP11-816F16-)", label)

    def test_zygosity(self):
        variant_descriptor = self._variant_interpretation.variation_descriptor
        allelic_state = variant_descriptor.allelic_state
        self.assertEqual("GENO:0000135", allelic_state.id)
        self.assertEqual("heterozygous", allelic_state.label)

    def test_structural_type(self):
        variant_descriptor = self._variant_interpretation.variation_descriptor
        structural_type = variant_descriptor.structural_type
        self.assertEqual("SO:1000029", structural_type.id)
        self.assertEqual("chromosomal_deletion", structural_type.label)



from .detailed_suppl_table import DetailedSupplTable
from .focus_count_table import FocusCountTable
from .hpoa_table_creator import HpoaTableCreator, HpoaTableBuilder
from .individual_table import IndividualTable
from .phenopacket_charts import PhenopacketCharts
from .phenopacket_ingestor import PhenopacketIngestor
from .phenopacket_table import PhenopacketTable
from .qc_visualizer import QcVisualizer
from .simple_age import SimpleAge
from .simple_patient import SimplePatient
from .simple_variant import SimpleVariant
from .hpo_category import HpoCategorySet





from .age_column_mapper import AgeColumnMapper
from .age_isoformater import AgeIsoFormater
from .age_of_death_mapper import AgeOfDeathColumnMapper
from .allelic_requirement import AllelicRequirement
from .case_encoder import CaseEncoder
from .case_template_encoder import CaseTemplateEncoder
from .citation import Citation
from .cohort_encoder import CohortEncoder
from .column_mapper import ColumnMapper
from .constant_column_mapper import ConstantColumnMapper
from .disease import Disease
from .disease_id_column_mapper import DiseaseIdColumnMapper
from .hgvs_variant import HgvsVariant
from .hpo_cr import HpoConceptRecognizer
from .hpo_exact_cr import HpoExactConceptRecognizer
from .hpo_parser import HpoParser
from .hp_term import HpTerm, HpTermBuilder
from .individual import Individual
from .metadata import MetaData
from .mixed_cohort_encoder import MixedCohortEncoder
from .option_column_mapper import OptionColumnMapper
from .sex_column_mapper import SexColumnMapper
from .simple_column_mapper import SimpleColumnMapper, SimpleColumnMapperGenerator
from .structural_variant import StructuralVariant
from .thresholded_column_mapper import ThresholdedColumnMapper
from .thresholder import Thresholder
from .variant import Variant
from .variant_column_mapper import VariantColumnMapper
from .variant_manager import VariantManager
from .variant_validator import VariantValidator



__all__ = [
    "AgeColumnMapper",
    "AgeIsoFormater",
    "AgeOfDeathColumnMapper",
    "AllelicRequirement",
    "CaseEncoder",
    "CaseTemplateEncoder",
    "Citation",
    "CohortEncoder",
    "ColumnMapper",
    "ConstantColumnMapper",
    "ColumnMapper",
    "Disease",
    "DiseaseIdColumnMapper",
    "HgvsVariant",
    "HpoConceptRecognizer",
    "HpoExactConceptRecognizer",
    "HpoParser",
    "HpTerm",
    "HpTermBuilder",
    "Individual",
    "MetaData",
    "MixedCohortEncoder",
    "OptionColumnMapper",
    "SexColumnMapper",
    "SimpleColumnMapper",
    "SimpleColumnMapperGenerator",
    "StructuralVariant",
    "ThresholdedColumnMapper",
    "Thresholder",
    "Variant",
    "VariantColumnMapper",
    "VariantManager",
    "VariantValidator",
]


from .disease import Disease

from typing import Dict


class DiseaseIdColumnMapper:

    def __init__(self, column_name, disease_id_map):
        """
        Constructor

        :param column_name: name of column with disease identifier
        :type column_name: str
        :param disease_id_map: dictionary with key, free text for disease identifier/name, value Disease object
        :type disease_id_map: Dict[str, Disease]
        """
        self._column_name = column_name
        self._disease_id_dict = disease_id_map


    def map_cell(self, cell_contents):
        """
        :param cell_contents: contents of a cell of the origiinal table
        :type cell_contents: str
        :returns: corresponding Disease object
        :rtype: Disease
        :raises: ValueError if the cell contents cannot be mapped
        """
        disease_id = str(cell_contents) # sometime pandas give us an int or an object
        if disease_id not in self._disease_id_dict:
            raise ValueError(f"Could not map disease \"{cell_contents}\"")
        return self._disease_id_dict.get(cell_contents)

    def get_column_name(self):
        return self._column_name


import os
import pickle
import pandas as pd
from typing import List, Dict
from collections import defaultdict
from .individual import Individual
from .variant_validator import VariantValidator
from .structural_variant import StructuralVariant

def get_pickle_filename(name):
    """
    provide standard filenaming convention. We pickle results from VariantValidator to avoid
    calling API multiple times in different runs. For instance, the pickled file of variants for
    the SCL4A1 cohort will be called "variant_validator_cache_SLC4A1.pickle"
    """
    return f"variant_validator_cache_{name}.pickle"

def load_variant_pickle(name):
    """
    Load the pickle file. If the file cannot be found, return None. If it is found, return the
    pickled object (which in our case will be a dictionary of Variant objects).
    """
    fname = get_pickle_filename(name)
    if not os.path.isfile(fname):
        return None
    # De-serialize the object from the file
    with open(fname, "rb") as f:
        loaded_object = pickle.load(f)
        return loaded_object

def write_variant_pickle(name, my_object):
    """
    Write a dictionary to pickled file
    :param my_object: In this application, this argument will be a dictionary of Variant objects).
    """
    fname = get_pickle_filename(name)
    with open(fname, "wb") as f:
        pickle.dump(my_object, f)



class VariantManager:
    """This class is designed to extract Variant objects from a pandas DataFrame that represents the input data.
    It will work out of the box for dataframes created by the CaseTemplateEncoder, and can be adapted to work
    with other datasets. The assumption is that there is one column per allele. For autosomal dominant, there would
    thus be one column. For recessive, there would be two columns. We do not try to automatically map non-HGVS (i.e.,
    chromosomal variants). Instead, we first map HGVS using VariantValidator, and then we return a Pandas DataFrame that
    shows the other variants. These can be use to create chromosomal deletions, duplications, and inversions. Finally,
    the class can be used to add variants to a list of Individual objects.

    :param df: DataFrame representing the input data
    :type df: pd.DataFrame
    :param individual_column_name: Name of the individual (patient) column
    :type individual_column_name: str
    :param transcript: accession number and version of the transcript, e.g., "NM_000342.3"
    :type transcript: str
    :param allele_1_column_name: name of the column with alleles (#1)
    :type allele_1_column_name: str
    :param allele_2_column_name: name of the column with alleles (#2), optional
    :type allele_2_column_name: str
    :param gene_symbol: Symbol of affectged gene (only required if chromosomal variants need to be coded)
    :type gene_symbol: str
    :param gene_id: HGNC identifier of affectged gene (only required if chromosomal variants need to be coded)
    :type gene_id: str
    """

    def __init__(self,
                df:pd.DataFrame,
                individual_column_name:str,
                transcript:str,
                gene_symbol:str,
                allele_1_column_name:str,
                allele_2_column_name:str=None,
                gene_id:str=None,
                overwrite:bool=False
                ):
        if not isinstance(df, pd.DataFrame):
            raise ValueError(f"The \"df\" argument must be a pandas DataFrame but was {type(df)}")
        if individual_column_name not in df.columns:
            raise ValueError(f"The \"individual_column_name\" argument must be a a column in df (a pandas DataFrame)")
        if allele_1_column_name not in df.columns:
            raise ValueError(f"The \"allele_1_column_name\" argument must be a a column in df (a pandas DataFrame)")
        if allele_2_column_name is not None and allele_2_column_name not in df.columns:
            raise ValueError(f"If not None, the \"allele_2_column_name\" argument must be a a column in df (a pandas DataFrame)")
        self._dataframe = df
        self._individual_column_name = individual_column_name
        #self._cohort_name = transcript
        self._transcript = transcript
        self._allele_1_column_name = allele_1_column_name
        self._allele_2_column_name = allele_2_column_name
        self._gene_symbol = gene_symbol
        self._gene_id = gene_id
        self._var_d = {}
        self._unmapped_alleles = set()
        self._individual_to_alleles_d = defaultdict(list)
        if "PMID" in df.columns:
            self._pmid_column_name = "PMID"
        else:
            self._pmid_column_name = None
        self._create_variant_d(overwrite)


    def _format_pmid_id(self, identifier, pmid):
        if pmid is not None:
            return f"{pmid}_{identifier}"
        else:
            return identifier

    def _get_identifier_with_pmid(self, row:pd.Series):
        """Get an identifier such as PMID_33087723_A2 for a daa row with PMID:33087723 and identifier within that publication A2

        Identifiers such as P1 are commonly used and there is a risk of a clash with collections of phenopackets from various papers.
        Therefore, we use an identifier such as PMID_33087723_A2 if we can find a PMID
        """
        individual_id = row[self._individual_column_name]
        individual_id = str(individual_id) # prevent automatic conversion into int for patient id 1, 2, 3 etc
        if self._pmid_column_name is not None:
            return self._format_pmid_id(identifier=individual_id, pmid=row[self._pmid_column_name])
        else:
            return individual_id

    def _create_variant_d(self, overwrite):
        """
        Creates a dictionary with all HGVS variants, and as a side effect creates a set with variants that
        are not HGVS and need to be mapped manually. This method has the following effects
        - self._var_d, a dictionary, is filled with key: HGVS strings, value: corresponding Variant objects
        - self._unmapped_alleles: set of all alleles that do not start eith "c." (non HGVS), that will need intervention by the user to map
        - self._individual_to_alleles_d: key individual ID, value-one or two element list of allele strings
        """
        if overwrite:
            v_d = {}
        else:
            v_d = load_variant_pickle(self._gene_symbol)
        if v_d is None:
            self._var_d = {}
        else:
            self._var_d = v_d
        genome_assembly = "hg38"  # Nothing else is good enough, sorry hg19
        variant_set = set()  # we expect HGVS nomenclature. Everything else will be parsed as chromosomal
        vvalidator = VariantValidator(genome_build=genome_assembly, transcript=self._transcript)
        # The DataFrame has two header rows.
        # For CaseTemplateEncoder, the second header row in effect is the first row of the DataFrame, so we drop it here.
        # For CaseTemplateEncoder, the second row will contain "str" in the second row of the PMID column
        # For other encoders, there may not be a "PMID" column, and if so it will not contain "str" in the second row
        if "PMID" in self._dataframe.columns and self._dataframe.iloc[0]["PMID"] == "str":
            df = self._dataframe.iloc[1:]
        else:
            df = self._dataframe
        for _, row in df.iterrows():
            individual_id = self._get_identifier_with_pmid(row=row)
            allele1 = row[self._allele_1_column_name]
            self._individual_to_alleles_d[individual_id].append(allele1)
            if allele1.startswith("c."):
                variant_set.add(allele1)
            else:
                self._unmapped_alleles.add(allele1)
            if self._allele_2_column_name is not None:
                allele2 = row[self._allele_2_column_name]
                if allele2 is None:
                    continue
                if allele2 == "na" or allele2 == "n/a":
                    continue
                self._individual_to_alleles_d[individual_id].append(allele2)
                if allele2.startswith("c."):
                    variant_set.add(allele2)
                else:
                    self._unmapped_alleles.add(allele2)
        for v in variant_set:
            if v in self._var_d:
                continue
            if v == "na" or v == "n/a":
                continue
            print(f"[INFO] encoding variant \"{v}\"")
            try:
                var = vvalidator.encode_hgvs(v)
                self._var_d[v] = var
            except Exception as e:
                print(f"[ERROR] Could not retrieve Variant Validator information for {v}: {str(e)}")
                self._unmapped_alleles.add(v) # This allows us to use the chromosomal mappers.
        write_variant_pickle(name=self._gene_symbol, my_object=self._var_d)

    def code_as_chromosomal_deletion(self, allele_set):
        """
        Code as Structural variants - chromosomal deletion (to be added to self._var_d)
        :param allele_set: Set of alleles (strings) for coding as Structural variants (chromosomal deletion)
        """
        # first check that all of the alleles are in self._unmapped_alleles
        if not allele_set.issubset(self._unmapped_alleles):
            raise ValueError("[ERROR] We can only map alleles that were passed to the constructor - are you trying to map \"new\" alleles?")
        if self._gene_id is None or self._gene_symbol is None:
            raise ValueError("[ERROR] We cannot use this method unless the gene ID (HGNC) and symbol were passed to the constructor")
        for allele in allele_set:
            var = StructuralVariant.chromosomal_deletion(cell_contents=allele, gene_symbol=self._gene_symbol, gene_id=self._gene_id)
            self._unmapped_alleles.remove(allele)
            self._var_d[allele] = var

    def code_as_chromosomal_duplication(self, allele_set):
        """
        Code as Structural variants - chromosomal duplication (to be added to self._var_d)
        :param allele_set: Set of alleles (strings) for coding as Structural variants (chromosomal duplication)
        """
        # first check that all of the alleles are in self._unmapped_alleles
        if not allele_set.issubset(self._unmapped_alleles):
            raise ValueError("[ERROR] We can only map alleles that were passed to the constructor - are you trying to map \"new\" alleles?")
        if self._gene_id is None or self._gene_symbol is None:
            raise ValueError("[ERROR] We cannot use this method unless the gene ID (HGNC) and symbol were passed to the constructor")
        for allele in allele_set:
            var = StructuralVariant.chromosomal_duplication(cell_contents=allele, gene_symbol=self._gene_symbol, gene_id=self._gene_id)
            self._unmapped_alleles.remove(allele)
            self._var_d[allele] = var

    def code_as_chromosomal_inversion(self, allele_set) -> None:
        """
        Code as Structural variants - chromosomal inversion (to be added to self._var_d)
        :param allele_set: Set of alleles (strings) for coding as Structural variants (chromosomal inversion)
        """
        # first check that all of the alleles are in self._unmapped_alleles
        if not allele_set.issubset(self._unmapped_alleles):
            raise ValueError("[ERROR] We can only map alleles that were passed to the constructor - are you trying to map \"new\" alleles?")
        if self._gene_id is None or self._gene_symbol is None:
            raise ValueError("[ERROR] We cannot use this method unless the gene ID (HGNC) and symbol were passed to the constructor")
        for allele in allele_set:
            var = StructuralVariant.chromosomal_inversion(cell_contents=allele, gene_symbol=self._gene_symbol, gene_id=self._gene_id)
            self._unmapped_alleles.remove(allele)
            self._var_d[allele] = var


    def to_summary(self) -> pd.DataFrame:
        """
        Create and return a DataFrame with current status of mapping efforts. The resulting dataframe can be used to
        check mapping and to retrieve the allele strings that could not be mapped (and therefore probably need to
        be mapped with one of the chromosomal methods)
        """
        dlist = []
        n_mapped = len(self._var_d)
        mapped_alleles = ", ".join(self._var_d.keys())
        d = {"status": "mapped", "count": n_mapped, "alleles": mapped_alleles}
        dlist.append(d)
        n_unmapped = len(self._unmapped_alleles)
        unmapped_alleles = ", ".join(self._unmapped_alleles)
        for allele in self._unmapped_alleles:
            if allele != allele.strip():
                unmapped_alleles = f"{unmapped_alleles}; warning \"{allele}\" has extra white space - check format"
        d = {"status": "unmapped", "count": n_unmapped, "alleles": unmapped_alleles}
        dlist.append(d)
        return pd.DataFrame(dlist)

    def add_variants_to_individuals(self, individual_list:List[Individual], hemizygous:bool=False):
        """
        Add Variant objects to individuals. Here, we use the map self._individual_to_alleles_d that
        relates the individual IDs to the allele strings in the original data, together with the
        map self._var_d, which relates the allele strings to Variant objects. If anything is missing, then
        we raise an Exception. Note currently this method cannot be used for X-dominant MOI.
        :param individual_list: list of Individuals to which we will add Variant objects
        :type individual_list: List[Individual]
        :param hemizygous: If True, this is an X-chromosomal gene and we assume hemizygous
        :type hemizygous: bool
        """
        if len(self._unmapped_alleles) > 0:
            raise ValueError(f"Need to map all allele strings before using this method but "
                            f"{len(self._unmapped_alleles)} were unmapped. Try variantManager.to_summary()")
        for i in individual_list:
            if i._citation is not None:
                pmid = i.get_citation().pmid
                individual_id = self._format_pmid_id(identifier=i.id, pmid=pmid)
            else:
                individual_id = self._format_pmid_id(identifier=i.id, pmid=None)
            if individual_id not in self._individual_to_alleles_d:
                raise ValueError(f"Did not find {i.id} in our dictionary of individuals. Note that this function is intended to work with CaseTemplateEncoder")
            vlist = self._individual_to_alleles_d.get(individual_id)
            if len(vlist) == 1:
                # assume monoallelic
                v = vlist[0]
                if v not in self._var_d:
                    raise ValueError(f"Could not find {v} in variant dictionary")
                var = self._var_d.get(v)
                if hemizygous:
                    var.set_hemizygous()
                else:
                    var.set_heterozygous()
                i.add_variant(var)
            elif len(vlist) == 2:
                if hemizygous:
                    raise ValueError("hemizygous argument set to True in present of two alleles")
                # assume biallelic
                v1 = vlist[0]
                v2 = vlist[1]
                if v1 not in self._var_d:
                    raise ValueError(f"Could not find {v1} in variant dictionary")
                if v2 not in self._var_d:
                    raise ValueError(f"Could not find {v2} in variant dictionary")
                if v1 == v2:
                    var = self._var_d.get(v1)
                    var.set_homozygous()
                    i.add_variant(var)
                else:
                    var1 = self._var_d.get(v1)
                    var1.set_heterozygous()
                    i.add_variant(var1)
                    var2 = self._var_d.get(v2)
                    var2.set_heterozygous()
                    i.add_variant(var2)

    def get_var(self, v):
        """Get a Variant object that corresponds to v.

        :param v: an HGVS string or free-text representation of a chromosomal variant
        :type v: str
        :returns: corresponding Variant
        :rtype: Variant
        """
        return self._var_d.get(v)

    def get_variant_d(self):
        """
        :returns:dictionary with key: original string for allele, value: Variant object
        :rtype: Dict[str, Variant]
        """
        return self._var_d








import unittest
import os
from src.pyphetools.creation import HpoParser

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')

class TestHpoParser(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls) -> None:
        parser = HpoParser(hpo_json_file=HP_JSON_FILENAME)
        cls.hpo_cr = parser.get_hpo_concept_recognizer()
        
    def test_not_none(self):
        self.assertIsNotNone(self.hpo_cr)

    def test_deprecated_nodes_not_parsed(self):
        """Test that the parser did not ingest a deprecated node"""
        # HP_0000284 is for the obsoleted term obsolete Abnormality of the ocular region
        self.assertFalse(self.hpo_cr.contains_term("HP0000284"))
        # HP:0000319 for for the valid term Smooth philtrum
        self.assertTrue(self.hpo_cr.contains_term("HP:0000319"))



from collections import defaultdict
from hpotk.constants.hpo.organ_system import *
from hpotk.model import TermId
from hpotk.ontology import Ontology
from hpotk.algorithm import exists_path
from typing import Dict



class HpoCategorySet:

    def __init__(self, ontology, organ_d = None) -> None:
        if not isinstance(ontology, Ontology):
            raise ValueError(f"ontology argument must be an hpo-toolkit Ontology object but was {type(ontology)}")
        self._ontology = ontology
        if organ_d is None:
            self._organ_d = self.get_default_organ_categories()
        else:
            self._organ_d = organ_d
        self._id_to_cat_d = defaultdict()
        


    def get_default_organ_categories(self) -> Dict:
        organ_d = defaultdict(TermId)
        organ_d['blood'] = ABNORMALITY_OF_BLOOD_AND_BLOOD_FORMING_TISSUES
        organ_d['genitourinary']  = ABNORMALITY_OF_GENITOURINARY_SYSTEM
        organ_d['head/neck'] = ABNORMALITY_OF_HEAD_OR_NECK
        organ_d['laboratory'] = ABNORMALITY_OF_METABOLISM_OR_HOMEOSTASIS
        organ_d['prenatal'] = ABNORMALITY_OF_PRENATAL_DEVELOPMENT_OR_BIRTH
        organ_d['breast'] = ABNORMALITY_OF_THE_BREAST
        organ_d['cardiovascular'] =  ABNORMALITY_OF_THE_CARDIOVASCULAR_SYSTEM
        organ_d['digestive'] = ABNORMALITY_OF_THE_DIGESTIVE_SYSTEM
        organ_d['ear'] = ABNORMALITY_OF_THE_EAR
        organ_d['endocrine'] = ABNORMALITY_OF_THE_ENDOCRINE_SYSTEM
        organ_d['eye'] =  ABNORMALITY_OF_THE_EYE
        organ_d['immune'] =  ABNORMALITY_OF_THE_IMMUNE_SYSTEM
        organ_d['integument'] = ABNORMALITY_OF_THE_INTEGUMENT
        organ_d['limbs'] =  ABNORMALITY_OF_THE_LIMBS
        organ_d['musculoskeletal'] = ABNORMALITY_OF_THE_MUSCULOSKELETAL_SYSTEM
        organ_d['nervous_system'] = ABNORMALITY_OF_THE_NERVOUS_SYSTEM
        organ_d['respiratory'] = ABNORMALITY_OF_THE_RESPIRATORY_SYSTEM
        organ_d['thoracic'] = ABNORMALITY_OF_THE_THORACIC_CAVITY
        organ_d['voice'] = ABNORMALITY_OF_THE_VOICE
        organ_d['cellular'] = ABNORMAL_CELLULAR_PHENOTYPE
        organ_d['constitutional'] = CONSTITUTIONAL_SYMPTOM
        organ_d['growth'] = GROWTH_ABNORMALITY
        organ_d['neoplasm'] = NEOPLASM
        return organ_d

    def get_category(self, termid):
        if isinstance(termid, str):
            hpo_term_id = TermId.from_curie(termid)
        elif isinstance(termid, TermId):
            hpo_term_id = termid.value
        else:
            raise ValueError(f"termid argument must be string (CURIE) or TermId object but was {type(termid)}")
        if hpo_term_id in self._id_to_cat_d:
            return self._id_to_cat_d.get(hpo_term_id)
        for cat, cat_hp_id in self._organ_d.items():
            if exists_path(self._ontology, hpo_term_id, cat_hp_id) or hpo_term_id == cat_hp_id:
                self._id_to_cat_d[hpo_term_id] = cat
                return cat
        print(f"Could not find category for {hpo_term_id}")
        return "not_found"


from google.protobuf.json_format import MessageToJson
import os
import pandas as pd
import phenopackets as PPKt
import re
from typing import Set
from collections import defaultdict
from .citation import Citation
from .column_mapper import ColumnMapper
from .constants import Constants
from .disease import Disease
from .hp_term import HpTerm
from .hpo_cr import HpoConceptRecognizer
from .individual import Individual
from .simple_column_mapper import SimpleColumnMapper
from .variant import Variant

ISO8601_REGEX = r"^P(\d+Y)?(\d+M)?(\d+D)?"


class CaseEncoder:
    """Encode a single case report in GA4GH Phenopacket format.

    This class should be used to encode a single case report; it outputs an Individual object that can be converted
    into a GA4GH phenopacket.

    :param hpo_cr: HpoConceptRecognizer for text mining
    :type hpo_cr: pyphetools.creation.HpoConceptRecognizer
    :param citation: PubMed identifierand title of this case report
    :type citation: Citation
    :param individual_id: Application specific individual identifier
    :type individual_id: str
    :param metadata: GA4GH MetaData object
    :type metadata: phenopackets.MetaData
    :param  age_at_last_exam: An ISO8601 string (e.g., P42Y2M) representing the age of the individual
    :type age_at_last_exam: str, optional
    :param sex: A string such as "male" or "FEMALE"
    :type sex: str
    :param disease: an object with the identifier and name of the disease diagnosis
    :type disease: Disease, optional
    """

    def __init__(self,
                hpo_cr: HpoConceptRecognizer,
                citation: Citation,
                individual_id:str,
                metadata:PPKt.MetaData,
                age_at_last_exam:str=None,
                sex:str=None,
                disease:Disease=None) -> None:
        if not isinstance(hpo_cr, HpoConceptRecognizer):
            raise ValueError(
                f"concept_recognizer argument must be HpoConceptRecognizer but was {type(hpo_cr)}")
        self._hpo_concept_recognizer = hpo_cr
        pmid = citation.pmid
        if not pmid.startswith("PMID:"):
            raise ValueError(f"Malformed pmid argument ({pmid}). Must start with PMID:")
        if age_at_last_exam is not None:
            match = re.search(ISO8601_REGEX, age_at_last_exam)
            if match:
                self._age_at_last_examination = age_at_last_exam
            else:
                raise ValueError(f"Could not parse {age_at_last_exam} as ISO8601 period")
        else:
            self._age_at_last_examination = None
        self._seen_hpo_terms = set() # Use to prevent duplicate HP annotations
        if not isinstance(metadata, PPKt.MetaData):
            raise ValueError(f"metadata argument must be phenopackets.MetaData but was {type(metadata)}")
        self._metadata = metadata
        female_symbols = {"f", "girl", "female", "woman", "w"}
        male_symbols = {"f", "boy", "male", "man", "m"}
        if sex is None:
            sex_symbol = Constants.UNKOWN_SEX_SYMBOL
        elif sex.lower() in female_symbols:
            sex_symbol = Constants.FEMALE_SYMBOL
        elif sex.lower() in male_symbols:
            sex_symbol = Constants.MALE_SYMBOL
        else:
            print(f"Warning: did not recognize sex symbol \"{sex}\"")
            sex_symbol = Constants.UNKOWN_SEX_SYMBOL
        ontology = hpo_cr.get_hpo_ontology()
        if ontology is None:
            raise ValueError("ontology cannot be None")
        self._individual = Individual(individual_id=individual_id, sex=sex_symbol)
        if age_at_last_exam is not None:
            self._individual.set_age(age_at_last_exam)
        if disease is not None:
            self._individual.set_disease(disease=disease)
        if citation is not None:
            self._individual.set_citation(citation=citation)


    def add_vignette(self, vignette, custom_d=None, custom_age=None, false_positive=None, excluded_terms:Set[str]=None) -> pd.DataFrame:
        """Add a description of a clinical encouter for text mining

        This method uses simple text mining to extract terms from the vignette. The optional custom_d argument can be
        used for HPO terms that are not mined by the text miner (users should use the preview to check the performance
        of the default miner and add terms as needed/desired). Note that HpTerms are added to the CaseEncoder object; these
        terms are also returned as a pandas DataFrame for checking.

        :param vignette: A free text description of a clinical encounter (we will use text mining to extract HPO terms)
        :type vignette: str
        :param custom_d: A dictionary of strings that correspond to HPO terms
        :type custom_d: dict
        :param custom_age: Age for the features in the current vignette
        :type custom_age: str
        :param false_positive: a set of strings that are omitted from text mining to avoid false positive results
        :type false_positive: set
        :param excluded_terms: similar to custom_d but for terms that are explicitly excluded in the vignette
        :type excluded_terms: Set[str]
        :returns: dataframe of (unique, alphabetically sorted) HpTerm objects with observed or excluded HPO terms
        :rtype: pd.DataFrame
        """
        if excluded_terms is None:
            excluded_terms = set()
        elif not isinstance(excluded_terms, set):
            raise ValueError(f"excluded_terms argument must be a set with HPO labels that have been explicitly excluded; but we got {type(excluded_terms)}")
        if false_positive is None:
            false_positive = set()
        elif not isinstance(false_positive, set):
            raise ValueError(f"false_positive argument must be a set with HPO labels that have been explicitly excluded; but we got {type(false_positive)}")
        if custom_d is None:
            custom_d = {}
        # replace new lines and multiple consecutive spaces with a single space
        text = re.sub(r'\s+', ' ', vignette.replace('\n', ' '))
        for fp in false_positive:
            text = text.replace(fp, " ")
        # results will be a list with HpTerm elements
        results = self._hpo_concept_recognizer.parse_cell(cell_contents=text, custom_d=custom_d)
        if custom_age is not None:
            current_age = custom_age
        elif self._age_at_last_examination is not None:
            current_age = self._age_at_last_examination
        else:
            current_age = Constants.NOT_PROVIDED
        for term in results:
            if term in self._seen_hpo_terms:
                continue
            else:
                self._seen_hpo_terms.add(term)
            if term.label in excluded_terms:
                term.excluded()
            term.set_onset(current_age)
            self._individual.add_hpo_term(term)
        return HpTerm.term_list_to_dataframe(results)

    def add_term(self, label=None, hpo_id=None, excluded=False, custom_age=None):
        """
        Add a single HPO term to the current case.
        :param label: Label of the HPO term
        :type label: str
        :param hpo_id: identifier of the term, e.g. HP:0001234
        :type hpo_id: str
        :param excluded: True iff the abnormally was explicitly excludeds
        :type excluded: bool
        :param custom_age: an ISO 8601 string representing the age of onset of the abnormality
        :type custom_age: str
        """
        if label is not None:
            results = self._hpo_concept_recognizer.parse_cell(cell_contents=label, custom_d={})
            if len(results) != 1:
                raise ValueError(f"Malformed label \"{label}\" we got {len(results)} results")
            hpo_term = results[0]
            hpo_term._observed = not excluded
            if custom_age is not None:
                hpo_term.set_onset(custom_age)
                self._individual.add_hpo_term(hpo_term)
            elif self._age_at_last_examination is not None:
                hpo_term.set_onset(self._age_at_last_examination)
                self._individual.add_hpo_term(hpo_term)
            else:
                hpo_term.set_onset(Constants.NOT_PROVIDED)
                self._individual.add_hpo_term(hpo_term)
            return HpTerm.term_list_to_dataframe([hpo_term])
        elif hpo_id is not None:
            hpo_term = self._hpo_concept_recognizer.get_term_from_id(hpo_id)
            if excluded:
                hpo_term._observed = not excluded
            if custom_age is not None:
                hpo_term.set_onset(custom_age)
                self._individual.add_hpo_term(hpo_term)
            elif self._age_at_last_examination is not None:
                hpo_term.set_onset(self._age_at_last_examination)
                self._individual.add_hpo_term(hpo_term)
            else:
                hpo_term.set_onset(Constants.NOT_PROVIDED)
                self._individual.add_hpo_term(hpo_term)
            return HpTerm.term_list_to_dataframe([hpo_term])
        else:
            return ValueError("Must call function with non-None value for either id or label argument")


    def add_variant_or_interpretation(self, variant_or_variantInterpretation):
        """
        :param variant_or_variantInterpretation: A presumed pathogenic variant observed in the current individual
        :type variant_or_variantInterpretation: Union[Variant, phenopackets.VariantInterpretation]
        """
        if isinstance(variant_or_variantInterpretation, Variant):
            interpretation = variant_or_variantInterpretation.to_ga4gh_variant_interpretation()
            self._individual.add_variant(interpretation)
        elif isinstance(variant_or_variantInterpretation, PPKt.VariantInterpretation):
            self._individual.add_variant(variant_or_variantInterpretation)
        else:
            raise ValueError(f"variant argument must be pyphetools Variant or GA4GH \
                phenopackets.VariantInterpretation but was {type(variant_or_variantInterpretation)}")

    def initialize_simple_column_maps(self, column_name_to_hpo_label_map, observed, excluded, non_measured=None):
        if observed is None or excluded is None:
            raise ValueError("Symbols for observed (e.g., +, Y, yes) and excluded (e.g., -, N, no) required")
        if not isinstance(column_name_to_hpo_label_map, dict):
            raise ValueError("column_name_to_hpo_label_map must be a dict with column to HPO label mappings")
        simple_mapper_d = defaultdict(ColumnMapper)
        for column_name, hpo_label in column_name_to_hpo_label_map.items():
            hp_term = self._hpo_concept_recognizer.get_term_from_label(hpo_label)
            mpr = SimpleColumnMapper(hpo_id=hp_term.id, hpo_label=hp_term.label, observed=observed, excluded=None,non_measured=non_measured)
            simple_mapper_d[column_name] = mpr
        return simple_mapper_d

    def get_hpo_term_dict(self):
        return self._annotations

    def get_individual(self)-> Individual:
        """
        :return: the pyphetools Individual object corresponding to the current case report
        :rtype: Individual
        """
        return self._individual

    def get_phenopacket(self):
        """
        :return: the GA4GH phenopacket corresponding to the current case report
        :rtype: PPKt.Phenopacket
        """
        individual = self.get_individual()
        pmid = individual.pmid
        phenopacket_id = pmid.replace(":", "_") + "_" + individual.id.replace(" ", "_").replace(":", "_")
        return individual.to_ga4gh_phenopacket(metadata=self._metadata, phenopacket_id=phenopacket_id)

    def output_phenopacket(self, outdir):
        """write a phenopacket to an output directory

        :param outdir: name of directory to write phenopackets
        :type outdir: str
        """
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        phenopacket = self.get_phenopacket()
        phenopacket_id = phenopacket.id
        json_string = MessageToJson(phenopacket)
        fname = phenopacket_id.replace(" ", "_") + ".json"
        outpth = os.path.join(outdir, fname)
        with open(outpth, "wt") as fh:
            fh.write(json_string)
            print(f"Wrote phenopacket to {outpth}")




import unittest
from src.pyphetools.creation import HpTerm, AllelicRequirement
from src.pyphetools.validation import ValidationResultBuilder



class TestValidationResult(unittest.TestCase):

    def test_redundant(self):
        """
        The function "redundant" in the builder should set the ErrorLevel to "WARNING"
        """
        term = HpTerm(hpo_id="HP:0000123", label="Fake label")
        term2 = HpTerm(hpo_id="HP:000987", label="Ancestor")
        validation_result = ValidationResultBuilder(
            phenopacket_id="id1"
        ).redundant_term(ancestor_term=term2, descendent_term=term).build()
        self.assertEqual("id1", validation_result.id)
        self.assertTrue(validation_result.is_warning())
        self.assertEqual("REDUNDANT", validation_result.category)
        term = validation_result.term
        self.assertIsNotNone(term)
        self.assertEqual("HP:000987", term.id)
        self.assertEqual("Ancestor", term.label)

    def test_not_measured(self):
        vresult = ValidationResultBuilder(
            phenopacket_id="id2"
        ).not_measured(HpTerm(hpo_id="HP:0012345", label="Something")).build()
        self.assertEqual("INFORMATION", vresult.error_level)
        self.assertEqual("NOT_MEASURED", vresult.category)

    def test_allele_count_mono_allelic(self):
        vresult = ValidationResultBuilder("id3").incorrect_allele_count(allelic_requirement=AllelicRequirement.MONO_ALLELIC, observed_alleles=2).build()
        self.assertEqual("ERROR", vresult.error_level)
        self.assertEqual("INCORRECT_ALLELE_COUNT", vresult.category)
        self.assertEqual("Expected one allele for monoallelic but got 2 alleles", vresult.message)

    def test_allele_count_bi_allelic(self):
        vresult = ValidationResultBuilder("id3").incorrect_allele_count(allelic_requirement=AllelicRequirement.BI_ALLELIC, observed_alleles=1).build()
        self.assertEqual("ERROR", vresult.error_level)
        self.assertEqual("INCORRECT_ALLELE_COUNT", vresult.category)
        self.assertEqual("Expected two alleles for biallelic but got 1 alleles", vresult.message)

    def test_insufficient_hpo(self):
        vresult = ValidationResultBuilder("id7").insufficient_hpos(min_hpo=2, n_hpo=1).build()
        self.assertEqual("ERROR", vresult.error_level)
        self.assertEqual("INSUFFICIENT_HPOS", vresult.category)
        self.assertEqual("Minimum HPO terms required 2 but only 1 found", vresult.message)

    def test_observed_and_excluded_is_unfixable(self):
        vresult = ValidationResultBuilder("id3").observed_and_excluded_term(term=HpTerm(hpo_id="HP:0012345", label="Something")).build()
        self.assertEqual("ERROR", vresult.error_level)
        self.assertEqual("OBSERVED_AND_EXCLUDED", vresult.category)
        self.assertTrue(vresult.is_unfixable_error())

    def test_conflict_is_not_unfixable(self):
        vresult = ValidationResultBuilder("id3").conflict(conflicting_term=HpTerm(hpo_id="HP:0012345", label="Something"),term=HpTerm(hpo_id="HP:0072345", label="Something else")).build()
        self.assertEqual("ERROR", vresult.error_level)
        self.assertEqual("CONFLICT", vresult.category)
        self.assertFalse(vresult.is_unfixable_error())





from typing import List
from ..creation.allelic_requirement import AllelicRequirement
from ..creation.individual import Individual
from .validated_individual import ValidatedIndividual
import hpotk

class CohortValidator:

    def __init__(self, cohort:List[Individual], ontology:hpotk.MinimalOntology, min_hpo:int,  allelic_requirement:AllelicRequirement=None) -> None:
        self._cohort = cohort
        self._ontology = ontology
        self._validated_individual_list = []
        for indi in cohort:
            vindi = ValidatedIndividual(individual=indi)
            vindi.validate(ontology=ontology, min_hpo=min_hpo, allelic_requirement=allelic_requirement)
            self._validated_individual_list.append(vindi)
        if len(cohort) != len(self._validated_individual_list):
            # should never happen
            raise ValueError(f"Invalid validation: size of cohort ={len(cohort)} but size of validated individual = {len(self._validated_individual_list)}")
        self._error_free_individuals = [vi.get_individual_with_clean_terms() for vi in self._validated_individual_list if not vi.has_unfixed_error()]
        self._v_individuals_with_unfixable_errors = [vi for vi in self._validated_individual_list if vi.has_unfixed_error()]

    def get_validated_individual_list(self):
        """
        :returns: list of all individuals with QC Validation results
        :rtype: List[ValidatedIndividual]
        """
        return self._validated_individual_list


    def get_error_free_individual_list(self) -> List[Individual]:
        """
        Returns a list of individuals from which the erroneous and redundant termas have been removed and from which individuals with errors (e.g., not enough HPO terms) have been removed.
        :returns: List of individuals with no errors
        :rtype: List[Individual]
        """
        return self._error_free_individuals

    def get_validated_individuals_with_unfixable_errors(self):
        """
        Returns a list of individuals with errors that cannot be automatically fixed.
        :returns: List of individuals with unfixable errors
        :rtype: List[ValidatedIndivudal]
        """
        return self._v_individuals_with_unfixable_errors


    def n_removed_individuals(self):
        return len(self._validated_individual_list) - len(self._error_free_individuals)

    def n_individuals(self):
        return len(self._validated_individual_list)

    def n_error_free_individuals(self):
        return len(self._error_free_individuals)

    def get_ontology(self):
        return self._ontology



from typing import List



class HtmlTableGenerator:
    """
    Helper class to generator an HTML table. This class is not intended to be used by client code.

    """

    def __init__(self, caption, header_items:List[str], rows:List[List[str]]) -> None:
        self._html_rows = []
        self._n_columns = len(header_items)
        self._html_rows.append('<table style="border: 2px solid black; align: "left">')
        self._html_rows.append(f'<caption>{caption}</caption>')
        self._html_rows.append(self._format_header(header_items=header_items))
        for row in rows:
            self._html_rows.append(self._format_row(row))
        self._html_rows.append('</table>') # close table content


    def _format_header(self, header_items):
        wrapped_items = [f"<th style=\"text-align: left;font-weight: bold;\">{x}</th>" for x in header_items]
        return "<tr>" + "".join(wrapped_items) + "</tr>"

    def _format_row(self, row:List[str]):
        if len(row) != self._n_columns:
            # should never happen if we construct the tables correctly
            raise ValueError(f"All rows need to have {self._n_columns} columns")
        wrapped_items = [f"<td style=\"text-align: left;\">{x}</td>" for x in row]
        return "<tr>" + "".join(wrapped_items) + "</tr>"

    def get_html(self):
        return "\n".join(self._html_rows)





import abc


class Variant(metaclass=abc.ABCMeta):
    """
    Superclass for classes that create GA4GH VariantInterpretationObjects.
    Subclass HgvsVariant is used for small variants encoded using HGVS notation, e.g., NM_00123.5:c.543G>T.
    Subclass StructuralVariant is used for structural variants that are likely to completely disrupt a gene.
    
    """
    
    def __init__(self):
        self._genotype = None

    @abc.abstractmethod
    def to_ga4gh_variant_interpretation(self, acmg=None):
        """
        Embed the variant object into a GA4GH Genomic Interpretation object (abstract method)

        The argument acmg must be one of the strings 'benign', 'likely_benign', 'uncertain_significance',
            'likely_pathogenic', or 'pathogenic' (underscores are optional). If it is not provided or
            not one of these strings, the level will be set to not available

        :param acmg: ACMG pathogenicity level
        :type acmg: str
        """
        pass
    
    def set_heterozygous(self):
        """
        Assign heterozygous allele status to this variant

        """
        self._genotype = 'heterozygous'
    
    def set_homozygous(self):
        """
        Assign homozygous allele status to this variant

        """
        self._genotype = 'homozygous'
        
    def set_hemizygous(self):
        """
        Assign hemizygous allele status to this variant

        """
        self._genotype = 'hemizygous'
        
    def set_genotype(self, gt):
        """
        Assign an allele status to this variant

        :param gt: The genotype (allele status) of this variant

        """
        genotype = gt.lower()
        if genotype not in {'heterozygous', 'homozygous', 'hemizygous'}:
            raise ValueError(f"Unrecognized genotype {gt}")
        self._genotype = genotype


from ..creation.allelic_requirement import AllelicRequirement
from ..creation.individual import Individual
from .content_validator import ContentValidator
from typing import List
from .validation_result import ValidationResult
from .ontology_qc import OntologyQC
import hpotk

class ValidatedIndividual:

    def __init__(self, individual:Individual) -> None:
        self._individual = individual
        self._clean_terms = []
        self._validation_errors = []


    def validate(self, ontology:hpotk.MinimalOntology, min_hpo:int, allelic_requirement:AllelicRequirement=None) -> None:
        """validate an Individual object for errors in the Ontology or the minimum number of HPO terms/alleles/variants

        :param ontology: HPO object
        :type ontology: hpotk.MinimalOntology
        :param min_hpo: minimum number of phenotypic features (HP terms) for this phenopacket to be considered valid
        :type min_hpo: int
        :param allelic_requirement: used to check number of alleles and variants
        :type allelic_requirement: AllelicRequirement
        """
        qc = OntologyQC(individual=self._individual, ontology=ontology)
        qc_validation_results = qc.get_error_list()
        self._validation_errors.extend(qc_validation_results)
        self._clean_terms = qc.get_clean_terms()
        self._individual.set_hpo_terms(self._clean_terms)
        cvalidator = ContentValidator(min_hpo=min_hpo, allelic_requirement=allelic_requirement)
        validation_results = cvalidator.validate_individual(individual=self._individual)
        self._validation_errors.extend(validation_results)
        # The following checks for remaining errors that would force us to remove the patient from the cohort
        self._unfixed_errors = [e  for e in self._validation_errors if e.is_unfixable_error()]


    def get_individual_with_clean_terms(self) -> Individual:
        """
        :returns: The Individual object input to the class from which redundant/conflicting terms have been removed
        :rtype: Individual
        """
        indi = self._individual
        indi.set_hpo_terms(self._clean_terms)
        return indi

    def get_validation_errors(self) -> List[ValidationResult]:
        return self._validation_errors

    def get_unfixed_errors(self)-> List[ValidationResult]:
        return self._unfixed_errors

    def n_errors(self) -> int:
        return len(self._validation_errors)

    def has_error(self) -> bool:
        """
        :returns: True iff this Individual had oneor more errors
        :rtype: bool
        """
        return len(self._validation_errors) > 0

    def has_unfixed_error(self) -> bool:
        """
        :returns: True iff this Individual had an error that we could not fix
        :rtype: bool
        """
        return len(self._unfixed_errors) > 0



import hpotk
import os
from typing import Dict
from .hpo_cr import HpoConceptRecognizer
from .hpo_exact_cr import HpoExactConceptRecognizer

HPO_JSON_URL = "http://purl.obolibrary.org/obo/hp/hp.json"

PHENOTYPE_ROOT = "HP:0000118"


class HpoParser:
    """
    Class to retrieve and parse the HPO JSON file using the HPO-Toolkit

    Users probably will want to pass the path to the hp.json file, but if not, the hp.json file
    will be downloadede each time the constructor is run

    :param hpo_json_file: path to the hp.json file
    :type hpo_json_file: str, optional
    """

    def __init__(self, hpo_json_file:str=None):
        if hpo_json_file is not None:
            if not os.path.isfile(hpo_json_file):
                raise FileNotFoundError(f"Could not find hp.json file at {hpo_json_file}")
            self._ontology = hpotk.load_ontology(hpo_json_file)
        else:
            self._ontology = hpotk.load_ontology(HPO_JSON_URL)

    def get_ontology(self):
        """
        :returns: a reference to the HPO
        :rtype: hpotk.Ontology
        """
        return self._ontology

    def get_label_to_id_map(self):
        """
        Create a map from a lower case version of HPO labels to the corresponding HPO id
        only include terms that are descendents of PHENOTYPE_ROOT

        :returns: a map from lower-case HPO term labels to HPO ids
        :rtype: Dict[str,str]
        """
        label_to_id_d = {}
        for t in self._ontology.terms:
            if t.is_obsolete:
                continue
            hpo_id = t.identifier.value
            if not self._ontology.graph.is_ancestor_of(PHENOTYPE_ROOT, hpo_id):
                continue
            label_to_id_d[t.name.lower()] = hpo_id
            # Add the labels of the synonyms
            if t.synonyms is not None and len(t.synonyms) > 0:
                for s in t.synonyms:
                    lc_syn = s.name.lower()
                    # only take synonyms with length at least 5 to avoid spurious matches
                    if len(lc_syn) > 4:
                        label_to_id_d[lc_syn] = hpo_id
        return label_to_id_d

    def get_id_to_label_map(self):
        """
        :returns: a map from HPO term ids to HPO labels
        :rtype: Dict[str,str]
        """
        id_to_label_d = {}
        for t in self._ontology.terms:
            if t.is_obsolete:
                continue
            id_to_label_d[t.identifier.value] = t.name
        return id_to_label_d

    def get_hpo_concept_recognizer(self) -> HpoConceptRecognizer:
        return HpoExactConceptRecognizer(label_to_id=self.get_label_to_id_map(),
                                         id_to_primary_label=self.get_id_to_label_map(),
                                         ontology=self.get_ontology())

    def get_version(self):
        return self._ontology.version.__str__()



from collections import defaultdict

import pandas as pd
from hpotk.model import TermId

from .hpo_category import HpoCategorySet

ALL_ROOT = TermId.from_curie("HP:0000001")
PHENOTYPIC_ABNORMALITY_ROOT = TermId.from_curie("HP:0000118")


class FocusCountTable:

    def __init__(self, patient_d, focus_id, ontology) -> None:
        if not isinstance(patient_d, dict):
            raise ValueError(f"patient_d argument must be dictionary but was {type(patient_d)}")
        self._patient_d = patient_d
        if isinstance(focus_id, str):
            self._focus_id_list = [focus_id]
        elif isinstance(focus_id, list):
            self._focus_id_list = focus_id
        else:
            raise ValueError(f"focus_id argument must be string or list of strings but was {type(focus_id)}")
        if len(self._focus_id_list) < 1:
            raise ValueError("Must provide at least one focus ID for FocusCountTable")
        self._ontology = ontology
        self._total_counts = defaultdict(int)
        self._focus_counts = defaultdict(int)
        self._non_focus_counts = defaultdict(int)
        self._total_counts_propagated = defaultdict(int)
        self._focus_counts_propagated = defaultdict(int)
        self._non_focus_counts_propagated = defaultdict(int)
        self._n_patients = len(patient_d)

        sex_d = defaultdict(int)
        var_d = defaultdict(int)

        for pat_id, pat in self._patient_d.items():
            sex_d[pat.get_sex()] += 1
            hpo_terms = pat.get_observed_hpo_d()
            anc_set = set()  # graph with ancestors induced by all terms of the patient
            for hp_id in hpo_terms.keys():
                # key is a string such as HP:0001234, value is an HpTerm object
                # we need to convert it to an object from hpo-toolkit because get_ancestors returns HpTerm objects
                hp_termid = TermId.from_curie(hp_id)
                ancs = self._ontology.graph.get_ancestors(hp_termid)
                anc_set.add(hp_termid)
                anc_set.update(ancs)
                if pat_id in focus_id:
                    self._focus_counts[hp_id] += 1
                else:
                    self._non_focus_counts[hp_id] += 1
                self._total_counts[hp_id] += 1
            for hp_id in anc_set:
                if hp_id == ALL_ROOT or hp_id == PHENOTYPIC_ABNORMALITY_ROOT:
                    continue
                if pat_id in focus_id:
                    self._focus_counts_propagated[hp_id.value] += 1
                else:
                    self._non_focus_counts_propagated[hp_id.value] += 1
                self._total_counts_propagated[hp_id.value] += 1
            variant_d = pat.get_variant_list()
            for var in variant_d:
                var_d[var] += 1

            # TODO figure out what to do with biallelic
        self._hpo_category_set = HpoCategorySet(ontology=ontology)

    def get_category(self, termid):
        # TODO figure out what to do with biallelic
        cat = self._hpo_category_set.get_category(termid=termid)
        if cat == "not_found":
            print(f"could not find category for {termid}")
        return cat

    def get_simple_table(self):
        """
        Get counts of terms within annotation propagation or thresholding
        """
        rows = []
        N = self._n_patients
        for hpid, total_count in self._total_counts.items():
            total_per = 100 * total_count / N
            total_s = f"{total_count}/{N} ({total_per:.1f}%)"
            hpterm = self._ontology.get_term(hpid)
            cat = self.get_category(termid=hpid)
            focus_count = self._focus_counts.get(hpid, 0)
            other_count = self._non_focus_counts.get(hpid, 0)
            d = {'category': cat, 'term': hpterm.name, 'HP:id': hpid, 'focus': focus_count, 'other': other_count,
                 'total': total_s, 'total_count': total_count}
            rows.append(d)
        df = pd.DataFrame(rows)
        df.set_index('category', inplace=True)
        return df.sort_values(['category', 'total_count'], ascending=[True, False])

    def get_thresholded_table(self, min_proportion: float = None, min_count: int = None):
        if min_count is None and min_proportion is None:
            raise ValueError("One of the arguments min_proportion and min_count must be provided")
        elif min_count is not None and min_proportion is not None:
            raise ValueError("Not more than one of the arguments min_proportion and min_count must be provided")
        elif min_proportion is not None:
            min_count = round(self._n_patients * min_proportion)
        print(f"Output terms with at least {min_count} counts")
        N = self._n_patients
        rows = []
        for hpid, total_count in self._total_counts_propagated.items():
            if total_count < min_count:
                continue
            total_per = 100 * total_count / N
            total_s = f"{total_count}/{N} ({total_per:.1f}%)"
            hpterm = self._ontology.get_term(hpid)
            cat = self.get_category(termid=hpid)
            focus_count = self._focus_counts_propagated.get(hpid, 0)
            other_count = self._non_focus_counts_propagated.get(hpid, 0)
            d = {'category': cat, 'term': hpterm.name, 'HP:id': hpid, 'focus': focus_count, 'other': other_count,
                 'total': total_s, 'total_count': total_count}
            rows.append(d)
        if len(rows) == 0:
            return pd.DataFrame(columns=['category', 'total_count'])
        df = pd.DataFrame(rows)
        df.set_index('category', inplace=True)
        return df.sort_values(['category', 'total_count'], ascending=[True, False])



from enum import IntEnum
from typing import List, Optional

from ..creation.hp_term import HpTerm
from ..creation.allelic_requirement import AllelicRequirement


class ErrorLevel(IntEnum):
    ERROR = 1
    WARNING = 2
    INFORMATION = 3
    UNKNOWN = 4



class Category(IntEnum):
    """
    These are the five types of error that we can identify.
    - REDUNDANT: A term and an ancestor of the term are both used to annotate the same individual
    - CONFLICT: A term is used to annotate an individual but an ancestor of the term is excluded in the same individual
    - INSUFFICIENT_HPOS: The number of HPO terms used to annotate an individual is less than the threshold value
    - INSUFFICIENT_ALLELES:  The number of alleles used to annotate an individual is less than the threshold value
    - INSUFFICIENT_VARIANTS:  The number of variants used to annotate an individual is less than the threshold value
    """
    REDUNDANT = 1
    CONFLICT = 2
    INSUFFICIENT_HPOS = 3
    INCORRECT_ALLELE_COUNT = 4
    INCORRECT_VARIANT_COUNT = 5
    MALFORMED_ID = 6
    MALFORMED_LABEL = 7
    NOT_MEASURED = 8
    OBSERVED_AND_EXCLUDED = 9
    DUPLICATE = 10
    UNKNOWN = 11


class ValidationResult:
    """
    A helper class to store the results of validation
    :param phenopacket_id: Identifier of the phenopacket being validated
    :type phenopacket_id: str
    :param message: description of the error/warning
    :type message: str
    :param errorlevel: whether this result is an error or a warning
    :type errorlevel: ErrorLevel
    :param category: type of QcError
    :type category: Category
    :param term: HpTerm that caused the error
    :type term: HpTerm

    """
    def __init__(self, phenopacket_id:str, message:str, errorlevel:ErrorLevel, category:Category, term:HpTerm=None):
        self._phenopacket_id = phenopacket_id
        self._message = message
        self._error_level = errorlevel
        self._category = category
        self._term = term

    @property
    def id(self):
        return self._phenopacket_id

    @property
    def message(self) -> str:
        """
        :returns: description of the cause of ValidationResult
        :rtype: str
        """
        return self._message

    @property
    def error_level(self)-> str:
        """
        :returns: the name of the ErrorLevel this ValidationResult is about
        :rtype: str
        """
        return self._error_level.name

    @property
    def term(self) -> Optional[HpTerm]:
        """
        :returns: A string representation of the HPO term this ValidationResult is about, if applicable, or empty string
        :rtype: Optional[str]
        """
        return self._term

    @property
    def category(self) -> str:
        """
        :returns: the name of the Category this ValidationResult is about
        :rtype: str
        """
        return self._category.name

    def is_error(self) -> bool:
        return self._error_level == ErrorLevel.ERROR

    def is_warning(self) -> bool:
        return self._error_level == ErrorLevel.WARNING

    def is_unfixable_error(self) -> bool:
        """Some errors cannot be fixed automatically and require manual attention.

        :returns: True iff this ValidationResult cannot be fixed automatically.
        :rtype: bool
        """
        return self._category in {Category.INSUFFICIENT_HPOS,
                                Category.INCORRECT_ALLELE_COUNT,
                                Category.INCORRECT_VARIANT_COUNT,
                                Category.MALFORMED_ID,
                                Category.MALFORMED_LABEL,
                                Category.OBSERVED_AND_EXCLUDED
                                }

    def get_items_as_array(self) -> List[str]:
        """
        :returns: A list of items (strings) intended for display
        :rtype: List[str]
        """
        if self._term is None:
            term = ""
        elif isinstance(self._term, HpTerm):
            term = self._term.hpo_term_and_id
        else:
            term = f"{self._term.name} ({self._term.identifier.value})"
        return [self.id, self.error_level, self.category, self.message, term]

    def __repr__(self):
        return f"{self._error_level}: {self._message}"


    @staticmethod
    def get_header_fields():
        return ["ID", "Level", "Category", "Message", "HPO Term"]




class ValidationResultBuilder:
    """
    This class is intended for internal use only, and makes constructing ValidatioResult objects a little easier.
    """

    def __init__(self, phenopacket_id:str):
        self._phenopacket_id = phenopacket_id
        self._error_level = ErrorLevel.UNKNOWN
        self._category = Category.UNKNOWN
        self._message = ""
        self._term = None

    def duplicate_term(self, redundant_term:HpTerm):
        """The HPO term is annotated as observed and excluded in the same individual. This is an unfixable error.

        :param redundant_term: HPO term that is annotated as observed and excluded
        :type redundant_term: HpTerm
        :returns: a reference to self so this command can be used as part of a builder.
        :rtype: ValidationResultBuilder
        """
        self._error_level = ErrorLevel.WARNING
        self._category = Category.DUPLICATE
        self._message = f"<b>{redundant_term.label}</b> is listed multiple times"
        self._term = redundant_term
        return self

    def observed_and_excluded_term(self, term:HpTerm):
        """The HPO term is annotated as observed and excluded in the same individual. This is an unfixable error.

        :param redundant_term: HPO term that is annotated as observed and excluded
        :type redundant_term: HpTerm
        :returns: a reference to self so this command can be used as part of a builder.
        :rtype: ValidationResultBuilder
        """
        self._error_level = ErrorLevel.ERROR
        self._category = Category.OBSERVED_AND_EXCLUDED
        self._message = f"Term {term.label} ({term.id}) was annotated to be both observed and excluded."
        self._term = term
        return self

    def redundant_term(self, ancestor_term:HpTerm, descendent_term:HpTerm):
        """The HPO term and one of its ancestors are both annotated as observed in the same individual.

        :param ancestor_term: Ancestor HPO term that is annotated as observed
        :type ancestor_term: HpTerm
        :param descendent_term: Descendent HPO term that is annotated as observed
        :type descendent_term: HpTerm
        :returns: a reference to self so this command can be used as part of a builder.
        :rtype: ValidationResultBuilder
        """
        self._error_level = ErrorLevel.WARNING
        self._category = Category.REDUNDANT
        self._message = f"<b>{ancestor_term.label}</b> is redundant because of <b>{descendent_term.label}</b>"
        self._term = ancestor_term
        return self

    def conflict(self, term:HpTerm, conflicting_term:HpTerm):
        message = f"{term.to_string()} conflicts with the excluded term {conflicting_term.to_string()} "
        self._error_level = ErrorLevel.ERROR
        self._category = Category.CONFLICT
        self._message = message
        self._term = conflicting_term
        return self

    def not_measured(self, term:HpTerm):
        self._error_level = ErrorLevel.INFORMATION
        self._category = Category.NOT_MEASURED
        self._term = term
        self._message = f"{term.hpo_term_and_id} was listed as not measured and will be omitted"
        return self

    def insufficient_hpos(self, min_hpo:int, n_hpo:int):
        self._message = f"Minimum HPO terms required {min_hpo} but only {n_hpo} found"
        self._error_level = ErrorLevel.ERROR
        self._category = Category.INSUFFICIENT_HPOS
        return self

    def incorrect_allele_count(self, allelic_requirement:AllelicRequirement, observed_alleles:int):
        if allelic_requirement == AllelicRequirement.MONO_ALLELIC:
            self._message = f"Expected one allele for monoallelic but got {observed_alleles} alleles"
        elif allelic_requirement == AllelicRequirement.BI_ALLELIC:
            self._message  = f"Expected two alleles for biallelic but got {observed_alleles} alleles"
        else:
            # should never happen
            raise ValueError("attempt to create incorrect_allele_count Error without defined allelic requirement")
        self._error_level = ErrorLevel.ERROR
        self._category = Category.INCORRECT_ALLELE_COUNT
        return self

    def incorrect_variant_count(self,allelic_requirement:AllelicRequirement, n_var:int):
        if allelic_requirement == AllelicRequirement.MONO_ALLELIC:
            self._message = f"Expected one variant for monoallelic but got {n_var} variants"
        elif allelic_requirement == AllelicRequirement.BI_ALLELIC:
            self._message  = f"Expected one or two variants for biallelic but got {n_var} variants"
        else:
            # should never happen
            raise ValueError("attempt to create incorrect_variant_count Error without defined allelic requirement")
        self._error_level = ErrorLevel.ERROR
        self._category = Category.INCORRECT_VARIANT_COUNT
        return self

    def set_message(self, msg):
        self._message = msg
        return self

    def malformed_hpo_id(self, malformed_term:HpTerm):
        self._error_level = ErrorLevel.ERROR
        self._category = Category.MALFORMED_ID
        self._message = f"Malformed term {malformed_term.label} with invalid HPO id {malformed_term.id}"
        return self

    def malformed_hpo_label(self, malformed_label, valid_term:HpTerm):
        self._error_level = ErrorLevel.ERROR
        self._category = Category.MALFORMED_LABEL
        self._message = f"Invalid label '{malformed_label}' found for {valid_term.name} ({valid_term.identifier.value})"
        self._term = valid_term
        return self

    def set_term(self, term:HpTerm):
        self._term = term
        return self

    def build(self) -> ValidationResult:
        return ValidationResult(phenopacket_id=self._phenopacket_id, message=self._message, errorlevel=self._error_level, category=self._category, term=self._term)







import os
import unittest

from src.pyphetools.creation import Individual, HpTerm, HpoParser
from src.pyphetools.validation import CohortValidator
from src.pyphetools.validation.validation_result import ErrorLevel, Category

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')

# The API requires us to pass a column name but the column name will not be used in the tests
TEST_COLUMN = "test"

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')

class TestCohortValidator(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        individual_A = Individual(individual_id="A")
        individual_A.add_hpo_term(HpTerm(hpo_id="HP:0000490", label="Deeply set eye"))
        individual_A.add_hpo_term(HpTerm(hpo_id="HP:0011525", label="Iris nevus"))
        individual_A.add_hpo_term(HpTerm(hpo_id="HP:0000490", label="Deeply set eye"))
        individual_B = Individual(individual_id="B")
        individual_B.add_hpo_term(HpTerm(hpo_id="HP:0000490", label="Deeply set eye"))
        individual_B.add_hpo_term(HpTerm(hpo_id="HP:0011525", label="Iris nevus"))
        cls._individual_A = individual_A
        cls._individual_B = individual_B
        parser = HpoParser(hpo_json_file=HP_JSON_FILENAME)
        cls.hpo_cr = parser.get_hpo_concept_recognizer()


    def test_redundant_term(self):
        cohort = [self._individual_A]
        hpo_ontology = self.hpo_cr.get_hpo_ontology()
        cvalidator = CohortValidator(cohort=cohort, ontology=hpo_ontology, min_hpo=1)
        validated_individuals = cvalidator.get_validated_individual_list()
        self.assertEqual(1, len(validated_individuals))
        vindindividualA = validated_individuals[0]
        self.assertTrue(vindindividualA.has_error())
        errors = vindindividualA.get_validation_errors()
        self.assertEqual(1, len(errors))
        error = errors[0]
        self.assertEqual("WARNING", error.error_level)
        self.assertEqual("DUPLICATE", error.category)
        self.assertEqual("<b>Deeply set eye</b> is listed multiple times", error.message)
        self.assertEqual("HP:0000490", error.term.id)
        self.assertEqual("Deeply set eye", error.term.label)

    def test_no_redundant_term(self):
        cohort = [self._individual_B]
        hpo_ontology = self.hpo_cr.get_hpo_ontology()
        cvalidator = CohortValidator(cohort=cohort, ontology=hpo_ontology, min_hpo=1)
        validated_individuals = cvalidator.get_validated_individual_list()
        self.assertEqual(1, len(validated_individuals))
        vindindividualB = validated_individuals[0]
        self.assertFalse(vindindividualB.has_error())

    def test_redundant_in_hierarchy(self):
        """
        0358596_patientA.json,ERROR,HpoAncestryValidator,Violation of the annotation propagation rule,
        "Phenotypic features of PMID_20358596_patient_A must not contain both an
        observed term (Bilateral facial palsy, HP:0430025) and an observed ancestor (Facial palsy, HP:0010628)"
        """
        individual_C = Individual(individual_id="C")
        individual_C.add_hpo_term(HpTerm(hpo_id="HP:0000490", label="Deeply set eye"))
        individual_C.add_hpo_term(HpTerm(hpo_id="HP:0011525", label="Iris nevus"))
        individual_C.add_hpo_term(HpTerm(hpo_id="HP:0430025", label="Bilateral facial palsy"))
        individual_C.add_hpo_term(HpTerm(hpo_id="HP:0010628", label="Facial palsy"))
        cohort = [individual_C]
        hpo_ontology = self.hpo_cr.get_hpo_ontology()
        cvalidator = CohortValidator(cohort=cohort, ontology=hpo_ontology, min_hpo=1)
        validated_individuals = cvalidator.get_validated_individual_list()
        self.assertEqual(1, len(validated_individuals))
        individual_C = validated_individuals[0]
        self.assertTrue(individual_C.has_error())
        errors = individual_C.get_validation_errors()
        self.assertEqual(1, len(errors))
        error = errors[0]
        self.assertEqual("<b>Facial palsy</b> is redundant because of <b>Bilateral facial palsy</b>", error.message)




import unittest
import os
from src.pyphetools.creation import HpoParser, OptionColumnMapper

HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')


class TestOptionMapper(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        parser = HpoParser(hpo_json_file=HP_JSON_FILENAME)
        cls.hpo_cr = parser.get_hpo_concept_recognizer()
        cls.severity_d = {'mild': 'Intellectual disability, mild',
                        'moderate': 'Intellectual disability, moderate',
                        'severe': 'Intellectual disability, severe'
                        }

    def test_hpo_cr_mild(self):
        """We should retrieve Intellectual disability, mild (HP:0001256)"""
        optionMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d=self.severity_d)
        results = optionMapper.map_cell("mild")
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual("HP:0001256", result.id)
        self.assertEqual("Intellectual disability, mild", result.label)

    def test_hpo_cr_moderate(self):
        """We should retrieve Intellectual disability, moderate (HP:0002342)"""
        optionMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d=self.severity_d)
        results = optionMapper.map_cell("moderate")
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual("HP:0002342", result.id)
        self.assertEqual("Intellectual disability, moderate", result.label)

    def test_hpo_cr_severe(self):
        """We should retrieve Intellectual disability, severe (HP:0010864) """
        optionMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d=self.severity_d)
        results = optionMapper.map_cell("severe")
        self.assertEqual(1, len(results))
        result = results[0]
        self.assertEqual("HP:0010864", result.id)
        self.assertEqual("Intellectual disability, severe", result.label)

    def test_with_acronyms(self):
        other_d = {
            "HP": "High palate",
            "D": "Dolichocephaly",
            "En": "Deeply set eye",  # i.e., Enophthalmus
            "DE": "Dural ectasia",
            "St": "Striae distensae"
        }
        otherMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d=other_d)
        res1 = otherMapper.map_cell("HP")
        res1.sort(key=lambda x: x.label) # sort alphabetically
        self.assertIsNotNone(res1)
        self.assertEqual(1, len(res1))
        hpterm = res1[0]
        self.assertEqual("High palate", hpterm.label)
        self.assertEqual("HP:0000218", hpterm.id)
        res2 = otherMapper.map_cell("HP,D")
        res2.sort(key=lambda x: x.label) # sort alphabetically
        self.assertIsNotNone(res2)
        self.assertEqual(2, len(res2))
        hpterm = res2[0]
        self.assertEqual("Dolichocephaly", hpterm.label)
        self.assertEqual("HP:0000268", hpterm.id)
        hpterm = res2[1]
        self.assertEqual("High palate", hpterm.label)
        self.assertEqual("HP:0000218", hpterm.id)

    def test_eso(self):
        oph_d = {"strabismus": "Strabismus",
                "esotropia": "Esotropia"}
        ophMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d=oph_d)
        res = ophMapper.map_cell("right sided esotropia")
        self.assertIsNotNone(res)
        self.assertEqual(1, len(res))
        hpterm = res[0]
        self.assertEqual("Esotropia", hpterm.label)
        self.assertEqual("HP:0000565", hpterm.id)

    def test_option_list(self):
        thumb_d = {"BT": "Broad thumb",
                    "BH": "Broad hallux",
                    "+": ["Broad thumb", "Broad hallux"]}
        thumbMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d=thumb_d)
        res = thumbMapper.map_cell("+")
        res.sort(key=lambda x: x.label) # sort alphabetically
        self.assertIsNotNone(res)
        self.assertEqual(2, len(res))
        hpterm0 = res[0]
        self.assertEqual("Broad hallux", hpterm0.label)
        self.assertEqual("HP:0010055", hpterm0.id)
        hpterm1 = res[1]
        self.assertEqual("Broad thumb", hpterm1.label)
        self.assertEqual("HP:0011304", hpterm1.id)

    def test_options_broad(self):
        thumb_d = {"BT": "Broad thumb",
                    "BH": "Broad hallux",
                    "+": [ "Broad thumb", "Broad hallux"]}
        thumbMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d=thumb_d)
        res = thumbMapper.map_cell("BT, BH")
        res.sort(key=lambda x: x.label) # sort alphabetically
        self.assertIsNotNone(res)
        self.assertEqual(2, len(res))
        hpterm0 = res[0]
        self.assertEqual("Broad hallux", hpterm0.label)
        self.assertEqual("HP:0010055", hpterm0.id)
        hpterm1 = res[1]
        self.assertEqual("Broad thumb", hpterm1.label)
        self.assertEqual("HP:0011304", hpterm1.id)

    def test_options_negative(self):
        thumb_d = {"BT": "Broad thumb",
                "BH": "Broad hallux",
                "+": ["Broad thumb", "Broad hallux"]}
        excluded_d = {"-":"Broad thumb"}
        thumbMapper = OptionColumnMapper(column_name="placeholder", concept_recognizer=self.hpo_cr, option_d=thumb_d,
                                        excluded_d=excluded_d)
        res = thumbMapper.map_cell("-")
        self.assertIsNotNone(res)
        self.assertEqual(1, len(res))
        hpterm0 = res[0]
        self.assertEqual("Broad thumb", hpterm0.label)
        self.assertEqual("HP:0011304", hpterm0.id)
        self.assertFalse(hpterm0.observed)


    def test_avoid_duplicates(self):
        """
        avoid adding the same term multiple times if one item is a substring of another
        """
        autistic_like_behavior_d = {
            'stereotypies': 'Motor stereotypy',
            'yes - was being investigated': 'Autism',
            'Severe autism': 'Autism',
            'yes': 'Autism'}
        exluded_d = {'no': 'Autism',
                    'no formal testing (no concern for ASD)': 'Autism', }
        autisticMapper = OptionColumnMapper(column_name="placeholder",
                                            concept_recognizer=self.hpo_cr,
                                            option_d=autistic_like_behavior_d,
                                            excluded_d=exluded_d)
        hpo_term_list = autisticMapper.map_cell("yes - was being investigated")
        self.assertEqual(1, len(hpo_term_list))

    def test_metabolic(self):
        """
        This test is for cases where the slash could be misinterpreted as a delimiter
        Now the code checks first for exact matches with the cell contents and the options_d
        """
        urine_xa_d = {'11.7umol/mmolCr': "Xanthinuria",
                    '1.7umol/mmolCr': "Xanthinuria"}
        urine_not_xa_d = {'0.04mmol/L': "Xanthinuria",
                        "1.6umol/mmolCr": "Xanthinuria",
                        "0.0214XA/Cr": "Xanthinuria",
                        "normal": "Xanthinuria"}
        urineXAmapper = OptionColumnMapper(column_name="placeholder",
                                        concept_recognizer=self.hpo_cr,
                                        option_d=urine_xa_d,
                                        excluded_d=urine_not_xa_d)
        hpo_term_list = urineXAmapper.map_cell('11.7umol/mmolCr')
        self.assertEqual(1, len(hpo_term_list))
        hpterm = hpo_term_list[0]
        self.assertEqual("Xanthinuria", hpterm.label)
        self.assertEqual("HP:0010934", hpterm.id)






import unittest
from src.pyphetools.creation import ThresholdedColumnMapper, Thresholder
from src.pyphetools.creation import HpTerm

class TestThresholdedColumnMapper(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        high = HpTerm(hpo_id="HP:0003155",label="Elevated circulating alkaline phosphatase concentration")
        low = HpTerm(hpo_id="HP:0003282",label="Low alkaline phosphatase")
        abn =  HpTerm(hpo_id="HP:0004379",label="Abnormality of alkaline phosphatase level")
        thresholder = Thresholder(hpo_term_high=high,
                                hpo_term_abn=abn,
                                hpo_term_low=low,
                                threshold_low=30,
                                threshold_high=120,
                                unit="U/L")
        #alkaline phosphatase concentration
        cls.mapper = ThresholdedColumnMapper(column_name="placeholder",
                                            thresholder=thresholder)

    def test_threshold_of_3(self):
        res = self.mapper.map_cell(3)
        self.assertIsNotNone(res)
        self.assertEqual(1, len(res))
        result = res[0]
        self.assertEqual("HP:0003282", result.id)
        self.assertEqual("Low alkaline phosphatase", result.label)
        self.assertTrue(result.observed)

    def test_threshold_of_400(self):
        res = self.mapper.map_cell(400)
        self.assertIsNotNone(res)
        self.assertEqual(1, len(res))
        result = res[0]
        self.assertEqual("HP:0003155", result.id)
        self.assertEqual("Elevated circulating alkaline phosphatase concentration", result.label)
        self.assertTrue(result.observed)

    def test_threshold_of_61(self):
        # normal result
        res = self.mapper.map_cell(61.1)
        self.assertIsNotNone(res)
        self.assertEqual(1, len(res))
        result = res[0]
        self.assertEqual("HP:0004379", result.id)
        self.assertEqual("Abnormality of alkaline phosphatase level", result.label)
        self.assertFalse(result.observed)

    def test_threshold_with_na(self):
        """_summary_
        Return cannot be measured if the threshold cannot be read
        """
        res = self.mapper.map_cell("NA")
        self.assertIsNotNone(res)
        self.assertEqual(1, len(res))
        result = res[0]
        self.assertEqual("HP:0004379", result.id)
        self.assertEqual("Abnormality of alkaline phosphatase level", result.label)
        self.assertFalse(result.measured)



import unittest
import os
from src.pyphetools.creation import HpoParser, HpTerm, Individual
from src.pyphetools.validation import OntologyQC


HP_JSON_FILENAME = os.path.join(os.path.dirname(__file__), 'data', 'hp.json')


class TestOntologyQC(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        parser = HpoParser(hpo_json_file=HP_JSON_FILENAME)
        hpo_cr = parser.get_hpo_concept_recognizer()
        cls._ontology = hpo_cr.get_hpo_ontology()
        individual = Individual(individual_id="id")
        cls._qc = OntologyQC(ontology=cls._ontology, individual=individual)

    def test_non_null(self):
        self.assertIsNotNone(self._qc)

    def test_conflict(self):
        """
        Arachnodactyly is a child of Slender finger, and so the following is a conflict
        """
        arachTerm = HpTerm(hpo_id="HP:0001166", label="Arachnodactyly", observed=True)
        slenderTerm = HpTerm(hpo_id="HP:0001238", label="Slender finger", observed=False)
        hpo_terms = [arachTerm, slenderTerm]
        individual = Individual(individual_id="id", hpo_terms=hpo_terms)
        qc = OntologyQC(ontology=self._ontology, individual=individual)
        qc_hpo_terms = qc.get_clean_terms()
        self.assertEqual(1, len(qc_hpo_terms))

    def test_do_not_detect_conflict_if_there_is_no_conflict(self):
        """
        These terms are unrelated so it is NOT a conflict
        """
        arachTerm = HpTerm(hpo_id="HP:0001166", label="Arachnodactyly", observed=True)
        # the following is not a conflict and should not be removed
        hipDislocation = HpTerm(hpo_id="HP:0002827", label="Hip dislocation", observed=False)
        hpo_terms = [arachTerm, hipDislocation]
        individual = Individual(individual_id="id", hpo_terms=hpo_terms)
        qc = OntologyQC(ontology=self._ontology, individual=individual)
        qc_hpo_terms = qc.get_clean_terms()
        self.assertEqual(2, len(qc_hpo_terms))

    def test_do_not_detect_conflict_if_there_is_no_conflict_2(self):
        """
        This is not a conflict because the subclass is excluded while the superclass is observed
        """
        arachTerm = HpTerm(hpo_id="HP:0001166", label="Arachnodactyly", observed=False)
        slenderTerm = HpTerm(hpo_id="HP:0001238", label="Slender finger", observed=True)
        hpo_terms = [arachTerm, slenderTerm]
        individual = Individual(individual_id="id", hpo_terms=hpo_terms)
        qc = OntologyQC(ontology=self._ontology, individual=individual)
        qc_hpo_terms = qc.get_clean_terms()
        self.assertEqual(2, len(qc_hpo_terms))

    def test_redundancy(self):
        arachTerm = HpTerm(hpo_id="HP:0001166", label="Arachnodactyly", observed=True)
        slenderTerm = HpTerm(hpo_id="HP:0001238", label="Slender finger", observed=True)
        hpo_terms = [arachTerm, slenderTerm]
        individual = Individual(individual_id="id", hpo_terms=hpo_terms)
        qc = OntologyQC(ontology=self._ontology, individual=individual)
        qc_hpo_terms = qc.get_clean_terms()
        self.assertEqual(1, len(qc_hpo_terms))

    def test_do_not_remove_if_not_redundant(self):
        arachTerm = HpTerm(hpo_id="HP:0001166", label="Arachnodactyly", observed=True)
        hipDislocation = HpTerm(hpo_id="HP:0002827", label="Hip dislocation", observed=False)
        hpo_terms = [arachTerm, hipDislocation]
        individual = Individual(individual_id="id", hpo_terms=hpo_terms)
        qc = OntologyQC(ontology=self._ontology, individual=individual)
        qc_hpo_terms = qc.get_clean_terms()
        self.assertEqual(2, len(qc_hpo_terms))


    def test_redundancy(self):
        arachTerm = HpTerm(hpo_id="HP:0001166", label="Arachnodactyly", observed=True)
        arachTerm2 = HpTerm(hpo_id="HP:0001166", label="Arachnodactyly", observed=True)
        hipDislocation = HpTerm(hpo_id="HP:0002827", label="Hip dislocation", observed=False)
        hpo_terms = [arachTerm, arachTerm2, hipDislocation]
        individual = Individual(individual_id="id", hpo_terms=hpo_terms)
        qc = OntologyQC(ontology=self._ontology, individual=individual)
        qc_hpo_terms = qc.get_clean_terms()
        self.assertEqual(2, len(qc_hpo_terms))
        error_list = qc.get_error_list()
        self.assertEqual(1, len(error_list))
        error = error_list[0]
        self.assertEqual("WARNING", error.error_level)
        self.assertEqual("DUPLICATE", error.category)


    def test_same_term_observed_and_excluded(self):
        arachTermObserved = HpTerm(hpo_id="HP:0001166", label="Arachnodactyly", observed=True)
        arachTermExcluded = HpTerm(hpo_id="HP:0001166", label="Arachnodactyly", observed=False)
        hipDislocation = HpTerm(hpo_id="HP:0002827", label="Hip dislocation", observed=True)
        hpo_terms = [arachTermObserved, hipDislocation, arachTermExcluded]
        individual = Individual(individual_id="id", hpo_terms=hpo_terms)
        qc = OntologyQC(ontology=self._ontology, individual=individual)
        error_list = qc.get_error_list()
        self.assertEqual(1, len(error_list))
        error = error_list[0]
        self.assertEqual("ERROR", error.error_level)
        self.assertEqual("OBSERVED_AND_EXCLUDED", error.category)
        self.assertEqual("Term Arachnodactyly (HP:0001166) was annotated to be both observed and excluded.", error.message)







import hpotk
from collections import Counter
from typing import List, Optional
from ..creation.hp_term import HpTerm
from .validation_result import ValidationResult, ValidationResultBuilder
from collections import defaultdict
from ..creation.individual import Individual



class OntologyQC:
    """
    This class performs three kind of checks/cleansing of ontology data
    1. negated superclass and observed subclass (this is an error in the original data)
    2. observed superclass and observed subclass (this is a redundancy but arguably not an error)
    3. Same term is excluded and observed (this is an unfixable error in the original data)

    """

    def __init__(self, ontology:hpotk.MinimalOntology, individual:Individual, fix_conflicts=True, fix_redundancies=True):
        self._ontology = ontology
        self._individual = individual
        self._phenopacket_id = individual.get_phenopacket_id()
        self._fix_conflict_flag = fix_conflicts
        self._fix_redundancy_flag = fix_redundancies
        self._errors = []
        self._clean_hpo_terms = self._clean_terms()


    def _fix_conflicts(self, observed_hpo_terms:List[HpTerm], excluded_hpo_terms) -> List[HpTerm]:
        """
        This class detects excluded superclasses that have observed subclasses -- a conflict.

        For instance, if an individual is annotated to the terms (1) excluded: Seizure [HP:0001250] and (2)
        observed - Clonic Seizure [HP:0020221], this is a conflict, because a person with clonic seizure also
        can be said to have seizure. Here, we assume that the excluded annotation is an error that we
        want to remove automatically and issue a warning. Thus, in this example, we would remove the
        annotation excluded: Seizure [HP:0001250], and in general the excluded superclass is removed
        if this kind of conflict is detected

        :param observed_hpo_terms: list of HPO terms (observed), can be empty
        :type observed_hpo_terms: List[HpTerm]
        :param excluded_hpo_terms: list of HPO terms (excluded), can be empty
        :type excluded_hpo_terms: List[HpTerm]
        :returns: the potentially cleansed list of excluded terms (the observed terms are never changed by this method
        :rtype: List[HpTerm]
        """
        if len(excluded_hpo_terms) == 0:
            # i.e., there can be no conflict
            return excluded_hpo_terms
        all_excluded_term_ids = {term.id for term in excluded_hpo_terms}
        conflicting_term_id_set = set()
        for term in observed_hpo_terms:
            for tid in all_excluded_term_ids:
                if term.id == tid:
                    # same term observed and excluded
                    # we cannot automatically fix this error
                    # this will be reported and the user will need to check the input data
                    error = ValidationResultBuilder(phenopacket_id=self._phenopacket_id).observed_and_excluded_term(term=term).build()
                    self._errors.append(error)
                elif self._ontology.graph.is_ancestor_of(tid, term.id):
                    conflicting_term_id_set.add(tid)
                    conflicting_term = self._ontology.get_term(term_id=tid)
                    cterm = HpTerm.from_hpo_tk_term(conflicting_term)
                    error = ValidationResultBuilder(phenopacket_id=self._phenopacket_id).conflict(term=term, conflicting_term=cterm).build()
                    self._errors.append(error)
        if len(conflicting_term_id_set) > 0:
            excluded_hpo_terms = [term for term in excluded_hpo_terms if term.id not in conflicting_term_id_set]
        return excluded_hpo_terms




    def _fix_redundancies(self, hpo_terms:List[HpTerm]) -> List[HpTerm]:
        """
        Remove redundant terms from a list of HPO terms.

        As a side effect, add a ValidationResult for each removed redundant term
        :param hpo_terms: original term list that might contain redundancies
        :type hpo_terms: List[HpTerm]
        :returns: list of HPO terms without redundancies
        :rtype hpo_terms: List[HpTerm]
        """
        all_terms = set(hpo_terms)
        # check for duplicates
        if len(all_terms) != len(hpo_terms):
            duplicates = [item for item, count in Counter(hpo_terms).items() if count > 1]
            for dup in duplicates:
                error = ValidationResultBuilder(self._phenopacket_id).duplicate_term(redundant_term=dup).build()
                self._errors.append(error)
            # The following removes duplicates under the assumption that all components of the HpTerm are equal
            hpo_terms = set(hpo_terms)
        # The following code checks for other kinds of redundancies
        redundant_term_d = {}
        for term in all_terms:
            for term2 in all_terms:
                # The ancestor, e.g. Seizure comes first, the other term, e.g. Clonic seizure, second
                # in the following function call
                if self._ontology.graph.is_ancestor_of(term2.id, term.id):
                    redundant_term_d[term2] = term
        # When we get here, we have scanned all terms for redundant ancestors
        non_redundant_terms = [ term for term in hpo_terms if term not in redundant_term_d]
        if len(redundant_term_d) > 0:
            for term, descendant in redundant_term_d.items():
                error = ValidationResultBuilder(self._phenopacket_id).redundant_term(ancestor_term=term, descendent_term=descendant).build()
                self._errors.append(error)
        return non_redundant_terms


    def _check_terms(self, hpo_terms:List[HpTerm]) -> None:
        for term in hpo_terms:
            hpo_id = term.id
            if not hpo_id in self._ontology:
                error = ValidationResultBuilder(self._phenopacket_id).malformed_hpo_id(malformed_term=term).build()
                self._errors.append(error)
            else:
                hpo_term = self._ontology.get_term(term_id=hpo_id)
                if hpo_term.name != term.label:
                    valid_term = HpTerm.from_hpo_tk_term(hpo_term)
                    error = ValidationResultBuilder(self._phenopacket_id).malformed_hpo_label(malformed_label=term.label,
                                                                                              valid_term=hpo_term).build()
                    self._errors.append(error)

    def _clean_terms(self) -> List[HpTerm]:
        """
        :returns: list of HPO terms without redundancies/conflicts
        :rtype hpo_terms: List[HpTerm]
        """
        by_age_dictionary =  defaultdict(list)
        for term in self._individual.hpo_terms:
            if not term.measured:
                self._errors.append(ValidationResultBuilder(self._phenopacket_id).not_measured(term=term).build())
            else:
                by_age_dictionary[term.onset].append(term)
        self._check_terms(self._individual.hpo_terms)
        clean_terms = []
        for onset, term_list in by_age_dictionary.items():
            observed_hpo_terms = [term for term in term_list if term.observed]
            excluded_hpo_terms = [term for term in term_list if not term.observed]
            if self._fix_redundancy_flag:
                observed_hpo_terms = self._fix_redundancies(observed_hpo_terms)
                excluded_hpo_terms = self._fix_redundancies(excluded_hpo_terms)
            if self._fix_conflict_flag:
                # this method checks and may fix the excluded terms (only)
                excluded_hpo_terms = self._fix_conflicts(observed_hpo_terms, excluded_hpo_terms)
            clean_terms.extend(observed_hpo_terms)
            clean_terms.extend(excluded_hpo_terms)
        return clean_terms

    def has_error(self) -> bool:
        """
        :returns: True iff errors were encountered
        :rtype: boolean
        """
        return len(self._errors) > 0

    def get_error_list(self) -> List[ValidationResult]:
        """
        :returns: a potential empty list of errors
        :rtype: List[str]
        """
        return self._errors

    def get_clean_terms(self) -> List[HpTerm]:
        return self._clean_hpo_terms


    def get_error_string(self) -> Optional[str]:
        """
        create and return a string that summarizes the redundancies and conflicts that were corrected

        :returns: a string summarizing errors or None if there were none
        :rtype: Optional[str]
        """
        if not self.has_error():
            return None
        redundancies = [e for e in self._errors if e.is_redundant()]
        conflicts = [e for e in self._errors if e.is_conflict()]
        e_string = ""
        if len(redundancies) > 0:
            red_terms = [e.hpo_term_and_id for e in redundancies]
            e_string = "The following redundant terms were removed: " + ", ".join(red_terms) + ". "
        if len(conflicts) > 0:
            conf_terms = [e.hpo_term_and_id for e in conflicts]
            e_string = e_string + "The following conflicting excluded terms were removed: " + ", ".join(conf_terms) + ". "
        return e_string


    @staticmethod
    def qc_cohort(individual_list:List[Individual]) -> List[Individual] :


        return individual_list



import os
from collections import defaultdict
import json
from google.protobuf.json_format import Parse
from .simple_patient import SimplePatient
from typing import Dict
import phenopackets as PPKt

class PhenopacketIngestor:
    """
    Ingest a collection of GA4GH Phenopacket objects from a directory

    :param indir: input directory
    :type indir: str
    :param recursive: Iff True, search subdirectorys for phenopackets
    :type recursive: bool, default False
    """

    def __init__(self, indir, recursive:bool=False) -> None:
        if not os.path.isdir(indir):
            raise ValueError(f"indir argument {indir} must be directory!")
        self._indir = indir
        self._phenopackets = []
        for file in os.listdir(indir):
            fname = os.path.join(indir, file)
            if fname.endswith(".json") and os.path.isfile(fname):
                with open(fname) as f:
                    data = f.read()
                    jsondata = json.loads(data)
                    ppack = Parse(json.dumps(jsondata), PPKt.Phenopacket())
                    self._phenopackets.append(ppack)
        print(f"[pyphetools] Ingested {len(self._phenopackets)} GA4GH phenopackets.")


    def get_simple_patient_dictionary(self) -> Dict:
        patient_d = defaultdict(SimplePatient)
        for ppack in self._phenopackets:
            patient = SimplePatient(ga4gh_phenopacket=ppack)
            patient_d[patient.get_subject_id()] = patient
        return patient_d


    def get_phenopacket_dictionary(self) -> Dict:
        patient_d = defaultdict(SimplePatient)
        for ppack in self._phenopackets:
            patient_d[ppack.id] = ppack
        return patient_d



import pandas as pd
import phenopackets as PPKt
from .constants import Constants
import hpotk

class HpTerm:
    """
    Class to represent a phenotypic observation as an HPO term with optional modifiers

    :param hpo_id: a Human Phenotype Ontology (HPO) identifier such as HP:0001166
    :type hpo_id: str
    :param label: The HPO label that corresponds to the id (note: This class does not check for correct match)
    :type label: str
    :param observed: a boolean that indicates whether the HPO term was observed (True) or excluded (False)
    :type observed: bool
    :param measured: a boolean that indicates whether the HPO was measured (True) or not explicitly measured (False)
    :type measured: bool
    :param onset: an ISO8601 string representing the age of onset, optional
    :type onset: str
    :param resolution: an ISO8601 string representing the age of resolution, optional
    :type resolution: str
    """
    def __init__(self, hpo_id, label, observed=True, measured=True, onset=Constants.NOT_PROVIDED, onset_term=None, resolution=None):
        if hpo_id is None or len(hpo_id) == 0 or not hpo_id.startswith("HP"):
            raise ValueError(f"invalid id argument: '{hpo_id}'")
        if label is None or len(label) == 0:
            raise ValueError(f"invalid label argument: '{label}'")
        self._id = hpo_id
        self._label = label
        self._observed = observed
        self._measured = measured
        if onset is None:
            self._onset = Constants.NOT_PROVIDED
        else:
            self._onset = onset
        self._onset_term = onset_term
        if resolution is None:
            self._resolution = Constants.NOT_PROVIDED
        else:
            self._resolution = resolution

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._id == other._id and self._label == other._label and self._measured == other._measured and self._onset == other._onset and self._resolution == other._resolution
        else:
            return NotImplemented

    def __hash__(self):
        return hash((self._id, self._label, self._observed, self._measured, self._onset, self._resolution))

    @property
    def id(self):
        """
        :returns: The HPO identifier, e.g., HP:0001166
        :rtype: str
        """
        return self._id

    @property
    def label(self):
        """
        :returns: The HPO label, e.g., Arachnodactyly
        :rtype: str
        """
        return self._label

    @property
    def observed(self):
        return self._observed

    @property
    def measured(self):
        """
        :returns: True iff a measurement to assess this abnormality (HpTerm) was performed
        :rtype: bool
        """
        return self._measured

    @property
    def onset(self):
        """
        :returns: iso8601 string representing the age this abnormality first was observed
        :rtype: str, optional
        """
        return self._onset

    def set_onset(self, onset):
        self._onset = onset

    @property
    def onset_term(self):
        return self._onset_term

    @property
    def resolution(self):
        """
        :returns: iso8601 string representing the age this abnormality resolved
        :rtype: str, optional
        """
        return self._resolution

    @property
    def display_value(self):
        """
        :returns: One of three strings describing the status of the term: "not measured", "excluded", or "observed"
        :rtype: str
        """
        if not self._measured:
            return "not measured"
        if not self._observed:
            return "excluded"
        else:
            return "observed"

    @property
    def hpo_term_and_id(self):
        """
        :returns: A string such as Arachnodactyly (HP:0001166) for display
        :rtype: str
        """
        return f"{self._label} ({self._id})"

    def _term_and_id_with_onset(self):
        if self._onset is not None and self._onset != Constants.NOT_PROVIDED:
            return f"{self.hpo_term_and_id}: onset {self._onset}"
        elif self._onset_term is not None:
            return f"{self.hpo_term_and_id}: onset {self._onset_term}"
        else:
            return self.hpo_term_and_id

    def __str__(self) -> str:
        if not self._measured:
            return f"not measured: {self._label} ({self._id})"
        elif not self._observed:
            return f"excluded: {self._term_and_id_with_onset()}"
        else:
            return self._term_and_id_with_onset()

    def to_string(self):
        return self.__str__()

    def excluded(self):
        """
        Sets the current term to excluded (i.e., the abnormality was sought but explicitly ruled out clinically)
        """
        self._observed = False

    def to_ga4gh_phenotypic_feature(self):
        """
        :returns: A GA4GH PhenotypcFeature corresponding to this HpTerm
        :rtype: phenopackets.PhenotypicFeature
        """
        pf = PPKt.PhenotypicFeature()
        pf.type.id = self._id
        pf.type.label = self._label
        if not self._observed:
            pf.excluded = True
        if self._onset != Constants.NOT_PROVIDED:
            pf.onset.age.iso8601duration = self._onset
        elif self.onset_term is not None:
            oclass = PPKt.OntologyClass()
            oclass.id = self._onset_term[1]
            oclass.label = self._onset_term[0]
            pf.onset.ontology_class.CopyFrom(oclass)
        if self._resolution != Constants.NOT_PROVIDED:
            pf.resolution.age.iso8601duration = self._resolution
        return pf


    @staticmethod
    def term_list_to_dataframe(hpo_list):
        if not isinstance(hpo_list, list):
            raise ValueError(f"hpo_list argument must be a list but was {type(hpo_list)}")
        if len(hpo_list) > 0:
            hpo1 = hpo_list[0]
            if not isinstance(hpo1, HpTerm):
                raise ValueError(f"hpo_list argument must consist of HpTerm objects but had {type(hpo1)}")
        if len(hpo_list) == 0:
            return pd.DataFrame(columns=['Col1', 'Col2', 'Col3'])
        items = []
        for hp in hpo_list:
            d = { "id": hp.id, "label": hp.label, "observed":hp.observed, "measured": hp.measured }
            items.append(d)
        return pd.DataFrame(items)

    @staticmethod
    def from_hpo_tk_term(hpotk_term:hpotk.Term):
        """Create a pyphetools HpTerm object from an hpo-toolkit Term object

        :param hpotk_term: A term from the HPO toolkit
        :type hpotk_term: hpotk.Term
        :returns: The corresponding HpTerm object
        :rtype: HpTerm
        """
        hpo_id = hpotk_term.identifier.value
        hpo_label = hpotk_term.name
        return HpTerm(hpo_id=hpo_id, label=hpo_label)


class HpTermBuilder:

    def __init__(self, hpo_id:str, hpo_label:str):
        if not hpo_id.startswith("HP:"):
            raise ValueError(f"Malformed HPO id {hpo_id}")
        if not len(hpo_id) == 10:
            length = len(hpo_id)
            raise ValueError(f"Malformed HPO id with length {len(hpo_id)}: {hpo_id}")
        self._hpo_id = hpo_id
        if hpo_label is None or len(hpo_label) < 3:
            raise ValueError(f"Malformed HPO label \"{hpo_label}\"")
        self._hpo_label = hpo_label
        self._observed=True
        self._measured=True
        self._onset=None
        self._onsetTerm=None

    def excluded(self):
        self._observed = False
        return self

    def not_measured(self):
        self._measured = False
        return self

    def iso8601_onset(self, onset):
        self._onset = onset
        return self

    def embryonal_onset(self):
        """Onset of disease at up to 8 weeks following fertilization (corresponding to 10 weeks of gestation).
        """
        self._onsetTerm = ["Embryonal onset", "HP:0011460"]
        return self

    def fetal_onset(self):
        """Onset prior to birth but after 8 weeks of embryonic development (corresponding to a gestational age of 10 weeks).
        """
        self._onsetTerm = ["Fetal onset", "HP:0011461"]
        return self

    def second_trimester_onset(self):
        """second trimester, which comprises the range of gestational ages from 14 0/7 weeks to 27 6/7 (inclusive)
        """
        self._onsetTerm = ["Second trimester onset", "HP:0034198"]
        return self

    def late_first_trimester_onset(self):
        """late first trimester during the early fetal period, which is defined as 11 0/7 to 13 6/7 weeks of gestation (inclusive).
        """
        self._onsetTerm = ["Late first trimester onset", "HP:0034199"]
        return self

    def third_trimester_onset(self):
        """third trimester, which is defined as 28 weeks and zero days (28+0) of gestation and beyond.
        """
        self._onsetTerm = ["Third trimester onset", "HP:0034197"]
        return self

    def antenatal_onset(self):
        """onset prior to birth
        """
        self._onsetTerm = ["Antenatal onset", "HP:0030674"]
        return self

    def congenital_onset(self):
        """A phenotypic abnormality that is present at birth.
        """
        self._onsetTerm = ["Congenital onset", "HP:0003577"]
        return self

    def neonatal_onset(self):
        """Onset of signs or symptoms of disease within the first 28 days of life.
        """
        self._onsetTerm = ["Neonatal onset", "HP:0003623"]
        return self


    def infantile_onset(self):
        """Onset of signs or symptoms of disease between 28 days to one year of life.
        """
        self._onsetTerm = ["Infantile onset", "HP:0003593"]
        return self

    def childhood_onset(self):
        """Onset of disease at the age of between 1 and 5 years.
        """
        self._onsetTerm = ["Childhood onset", "HP:0011463"]
        return self

    def juvenile_onset(self):
        """Onset of signs or symptoms of disease between the age of 5 and 15 years.
        """
        self._onsetTerm = ["Juvenile onset", "HP:0003621"]
        return self

    def young_adult_onset(self):
        """Onset of disease at the age of between 16 and 40 years.
        """
        self._onsetTerm = ["Young adult onset", "HP:0011462"]
        return self

    def middle_age_onset(self):
        """onset of symptoms at the age of 40 to 60 years.
        """
        self._onsetTerm = ["Middle age onset", "HP:0003596"]
        return self

    def late_onset(self):
        """Onset of symptoms after the age of 60 years.
        """
        self._onsetTerm = ["Late onset", "HP:0003584"]
        return self

    def build(self) -> HpTerm:
        if self._onset is not None and self._onsetTerm is not None:
            raise ValueError("Both onset and onsetTerm were set but only a maximum of one of them may be not None")
        return HpTerm(hpo_id=self._hpo_id,
                    label=self._hpo_label,
                    observed=self._observed,
                    measured=self._measured,
                    onset=self._onset,
                    onset_term=self._onsetTerm)





import hpotk

from .hpo_cr import HpoConceptRecognizer
from .hp_term import HpTerm
from .column_mapper import ColumnMapper
from .simple_column_mapper import SimpleColumnMapper

import re
from typing import List, Union
from collections import defaultdict


class ConceptMatch:
    def __init__(self, term, start: int, end: int) -> None:
        self._hp_term = term
        self._start = start
        self._end = end

    def length(self):
        return 1 + (self._end - self._start)

    @property
    def label(self):
        return self._hp_term.label

    @property
    def tid(self):
        return self._hp_term.id

    @property
    def term(self):
        return self._hp_term

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    def overlaps(self, other):
        if self.end >= other.start >= self.start:
            return True
        elif self.end >= other.end >= self.start:
            return True
        else:
            return False


class HpoExactConceptRecognizer(HpoConceptRecognizer):
    # self._label_to_id_d, self._id_to_primary_label
    def __init__(self, label_to_id, id_to_primary_label, ontology:hpotk.Ontology=None):
        super().__init__()
        if not isinstance(label_to_id, dict):
            raise ValueError("label_to_id_d argument must be dictionary")
        if not isinstance(id_to_primary_label, dict):
            raise ValueError("labels_to_primary_label_d argument must be dictionary")
        self._id_to_primary_label = id_to_primary_label
        self._label_to_id = label_to_id
        self._ontology = ontology

    def parse_cell(self, cell_contents, custom_d=None) -> List[HpTerm]:
        """parse the contents of one table cell

        Args:
            cell_contents (str): description of patient phenotypes. Assumption is that if there are new lines, any given phenotype is on its own line
            custom_d (dict, optional): User-provided dictionary with mappings to HPO terms. Defaults to None.
        """
        if not isinstance(cell_contents, str):
            print(
                f"Error: cell_contents argument ({cell_contents}) must be string but was {type(cell_contents)} -- coerced to string")
            cell_contents = str(cell_contents)

        # lines = self._split_into_lines(cell_contents)
        cell_text = cell_contents.replace("\n", " ")
        if custom_d is None:
            # initialize to empty dictionary if this argument is not passed
            # to avoid needed to check for None in other functions
            custom_d = defaultdict()
        return self._parse_contents(cell_text=cell_text, custom_d=custom_d)


    def _get_exact_match_in_custom_d(self, cell_text, custom_d) -> List[HpTerm]:
        """This method is called by _parse_contents if cell_text was present in custom_d

        If the match does not result in at least one HpTerm match, it is an error and the custom_d probably has an incorrect string

        :param cell_text: The text of a table cell
        :type cell_text: str
        :param custom_d: key - text in original table value-corresponding HPO label
        :type custom_d: Dict[str, Union[str,List[str]]]
        :returns: list of HPO terms
        :rtype: List[HpTerm]
        """
        label = custom_d.get(cell_text)
        results = []
        if isinstance(label, list):
            for lab in label:
                hp_term = self.get_term_from_label(label=lab)
                if hp_term is None:
                    raise ValueError(f"Could not get HpTerm from {lab}")
                else:
                    results.append(hp_term)
        else:
            hp_term = self.get_term_from_label(label=label)
            if hp_term is None:
                    raise ValueError(f"Could not get HpTerm from {label}")
            else:
                results.append(hp_term)
        return results


    def _find_text_within_custom_items(self, lc_chunk, custom_d) -> List[HpTerm]:
        hits = []
        # Note that chunk has been stripped of whitespace and lower-cased already
        for original_text, hpo_label in custom_d.items():
            lc_original = original_text.lower()
            startpos = lc_chunk.find(lc_original)
            if startpos < 0:
                continue
            endpos = startpos + len(lc_original) - 1
            if isinstance(hpo_label, str):
                hp_term = self.get_term_from_label(hpo_label)
                hits.append(ConceptMatch(term=hp_term, start=startpos, end=endpos))
            elif isinstance(hpo_label, list):
                for h in hpo_label:
                    hp_term = self.get_term_from_label(h)
                    hits.append(ConceptMatch(term=hp_term, start=startpos, end=endpos))
        return hits

    def _find_hpo_term_in_lc_chunk(self, lc_chunk) -> List[HpTerm]:
        hits = []
        for lower_case_hp_label, hpo_tid in self._label_to_id.items():
                key = lower_case_hp_label.lower()
                startpos = lc_chunk.find(key)
                endpos = startpos + len(key) - 1
                if startpos < 0:
                    continue
                # If we get here, we demand that the match is a complete word
                # This is because otherwise we get some spurious matches such as Pica HP:0011856 matching to typical
                # Create a regex to enforce the match is at word boundary
                BOUNDARY_REGEX = re.compile(r'\b%s\b' % key, re.I)
                if BOUNDARY_REGEX.search(lc_chunk):
                    hp_term = self.get_term_from_id(hpo_id=hpo_tid)  # Get properly capitalized label
                    hits.append(ConceptMatch(term=hp_term, start=startpos, end=endpos))
        return hits

    def _parse_contents(self, cell_text, custom_d) -> List[HpTerm]:
        """Parse the contents of a cell for HPO terms
        Args:
            cell_text (str): The text of a table cell
            custom_d (dict): key - text in original table value-corresponding HPO label
        """
        if cell_text in custom_d:
            return self._get_exact_match_in_custom_d(cell_text=cell_text, custom_d=custom_d)
        chunks = self._split_line_into_chunks(cell_text)
        results = []
        for chunk in chunks:
            lc_chunk = chunk.lower()
            hits_1 = self._find_text_within_custom_items(lc_chunk=lc_chunk, custom_d=custom_d)
            hits_2 = self._find_hpo_term_in_lc_chunk(lc_chunk=lc_chunk)
            hits_1.extend(hits_2)
            results.extend(self._get_non_overlapping_matches(hits=hits_1))
        return results

    def parse_cell_for_exact_matches(self, cell_text, custom_d) -> List[HpTerm]:
        """
        Identify HPO Terms from the contents of a cell whose label exactly matches a string in the custom dictionary

        :param cell_contents: a cell of the original table
        :type cell_contents: str
        :param custom_d: a dictionary with keys for strings in the original table and their mappings to HPO labels
        :type custom_d: Dict[str,str]
        """
        if cell_text in custom_d:
            return self._get_exact_match_in_custom_d(cell_text=cell_text, custom_d=custom_d)
        chunks = self._split_line_into_chunks(cell_text)
        results = []
        for chunk in chunks:
            lc_chunk = chunk.lower()
            hits = self._find_text_within_custom_items(lc_chunk=lc_chunk, custom_d=custom_d)
            results.extend(self._get_non_overlapping_matches(hits=hits))
        return

    def _get_non_overlapping_matches(self, hits:[List[ConceptMatch]]) -> List[HpTerm]:
        """The prupose of this method is to choose a list of non-overlapping matches

        Sometimes, we get multiple matches that partially overlap. We will greedily take the longest matches and discard overlaps.

        :param hits: list of ConceptMatch objects that encode HpTerm matches and their positions
        :type hits: List[ConceptMatch]
        :returns: a list of non-overlapping HtTerm objects (matches)
        :rtype: List[HpTerm]
        """
        sorted_hits = sorted(hits, key=ConceptMatch.length, reverse=True)
        # Choose longest hits first and skip hits that overlap with previously chosen hits
        chosen_hits = set()
        for hit in sorted_hits:
            keeper = True
            for ch in chosen_hits:
                if ch.overlaps(hit):
                    keeper = False
                    break
            if keeper:
                chosen_hits.add(hit)
        results = []
        for ch in chosen_hits:
            results.append(ch.term)
        return results

    def _split_line_into_chunks(self, line):
        """Split a line into chunks and remove white space from beginning and end of each chunk

        Args:
            line (str): one line of a potentially multi-line Table cell.
        """
        delimiters = ',;|/'
        regex_pattern = '|'.join(map(re.escape, delimiters))
        chunks = re.split(regex_pattern, line)
        return [chunk.strip().lower() for chunk in chunks]

    def get_term_from_id(self, hpo_id) -> HpTerm:
        if not hpo_id.startswith("HP:"):
            raise ValueError(f"Malformed HP id '{hpo_id}' - must start with HP:")
        if hpo_id not in self._id_to_primary_label:
            raise ValueError(f"Could not find id {hpo_id} in dictionary")
        label = self._id_to_primary_label.get(hpo_id)
        return HpTerm(hpo_id=hpo_id, label=label)

    def get_term_from_label(self, label) -> HpTerm:
        label_lc = label.lower()  # the dictionary was constructed in lower case!
        if label_lc not in self._label_to_id:
            raise ValueError(f"Could not find HPO id for {label}")
        hpo_id = self._label_to_id.get(label_lc)
        return HpTerm(hpo_id=hpo_id, label=label)

    def contains_term(self, hpo_id) -> bool:
        return hpo_id in self._id_to_primary_label

    def contains_term_label(self, hpo_label) -> bool:
        """return True iff the argument is the primary label of an HPO term
        """
        return hpo_label.lower() in self._label_to_id

    def initialize_simple_column_maps(self, column_name_to_hpo_label_map, observed, excluded):
        if observed is None or excluded is None:
            raise ValueError("Symbols for observed (e.g., +, Y, yes) and excluded (e.g., -, N, no) required")
        if not isinstance(column_name_to_hpo_label_map, dict):
            raise ValueError("column_name_to_hpo_label_map must be a dict with column to HPO label mappings")
        simple_mapper_d = defaultdict(ColumnMapper)
        for column_name, hpo_label_and_id in column_name_to_hpo_label_map.items():
            if not isinstance(hpo_label_and_id, list):
                raise ValueError(f"Expected {hpo_label_and_id} to be a two-item list with HPO label and id")
            hpo_label = hpo_label_and_id[0]
            expected_id = hpo_label_and_id[1]
            hp_term = self.get_term_from_label(hpo_label)
            if hp_term.id != expected_id:
                raise ValueError(f"Got {hp_term.id} but was expecting {expected_id} for {hpo_label}")
            mpr = SimpleColumnMapper(column_name=column_name, hpo_id=hp_term.id, hpo_label=hp_term.label, observed=observed, excluded=excluded)
            simple_mapper_d[column_name] = mpr
        return simple_mapper_d

    def get_hpo_ontology(self):
        """
        :returns: a reference to the HPO-toolkit Ontology object for the HPO
        :rtype: hpotk.MinimalOntology
        """
        return self._ontology



from ..creation.hp_term import HpTerm


class SimpleAge:

    def __init__(self, age_string) -> None:
        self._age_string = age_string
        self._gestational_weeks = None
        self._gestational_days = None
        self._years = None
        self._months = None
        self._weeks = None
        self._days = None

        if age_string.startswith("P"):
            self._parse_iso8601()
        elif age_string.startswith("GA"):
            self._parse_gestational_age()
        else:
            raise ValueError(f"Malformed age string (must be iso8601 or gestational age) : {age_string}")


    def _parse_gestational_age():
        raise ValueError("not implemented yet")


    def _parse_iso8601(self):
        iso8601 = self._age_string[1:]
        self._days = 0
        if "Y" in iso8601:
            i = iso8601.index("Y")
            self._years = int(iso8601[:i])
            iso8601 = iso8601[(1+i):]
        else:
            self._years = 0
        if "M" in iso8601:
            i = iso8601.index("M")
            self._months = int(iso8601[:i])
            iso8601 = iso8601[(1+i):]
        else:
            self._months = 0
        if "W" in iso8601:
            # don't keep weeks; convert to days
            i = iso8601.index("W")
            self._days +=  7 * int(iso8601[:i])
            iso8601 = iso8601[(1+i):]
        else:
            self._weeks = 0
        if "D" in iso8601:
            i = iso8601.index("D")
            self._days += int(iso8601[:i])
        # for sorting, we will use the following expression for total days
        t = 365.25 * self._years + 30.437 * self._months + self._days
        self._total_days = int(t)


    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._age_string == other._age_string
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self._age_string)

    def __str__(self) -> str:
        return self._age_string

    def get_total_days(self) -> int:
        """get total number of days of life. This simplifies sorting.
        """
        return self._total_days

    def to_hpo_onset_term(self):
        if self._years > 60:
            return HpTerm(hpo_id="HP:0003584", label="Late onset")
        if self._years >= 40:
            return HpTerm(hpo_id="HP:0003596", label="Middle age onset")
        if self._years >= 25:
            return HpTerm(hpo_id="HP:0025710", label="Late young adult onset")
        if self._years >= 19:
            return HpTerm(hpo_id="HP:0025709", label="Intermediate young adult onset")
        if self._years >= 16:
            return HpTerm(hpo_id="HP:0011462", label="Early young adult onset")
        if self._years > 5 or (self._years==5 and self._months > 0):
            return HpTerm(hpo_id="HP:0003621", label="Juvenile onset")
        if self._years >= 1:
            return HpTerm(hpo_id="HP:0011463", label="Childhood onset")
        if self._months >= 1:
            return HpTerm(hpo_id="HP:0003593", label="Infantile onset")
        if self._days > 1:
            return HpTerm(hpo_id="HP:0003623", label="Neonatal onset")
        if self._days == 0 or self._days == 1:
            return HpTerm(hpo_id="HP:0003577", label="Congenital onset")
        raise ValueError(f"Could not find onset for {self._age_string}")

